"""
Indices Helper Functions Submodule
==================================

Functions that encapsulate logic and can be shared by many indices,
but are not particularly index-like themselves (those should go in the :py:mod:`xclim.indices.generic` module).
"""

from __future__ import annotations

from collections.abc import Callable, Mapping
from datetime import timedelta
from inspect import stack
from typing import Any, Literal, cast

import cf_xarray  # noqa: F401, pylint: disable=unused-import
import cftime
import numba as nb
import numpy as np
import xarray as xr
from xarray import CFTimeIndex

try:
    from xarray.coding.calendar_ops import _datetime_to_decimal_year
except ImportError:
    XR2409 = True
else:
    XR2409 = False

try:
    from flox.xarray import rechunk_for_blockwise

    flox_err = None
except ImportError:
    rechunk_for_blockwise = None

from xclim.core import DayOfYearStr, Quantified
from xclim.core.calendar import ensure_cftime_array, get_calendar, parse_offset, select_time
from xclim.core.options import MAP_BLOCKS, OPTIONS
from xclim.core.units import convert_units_to
from xclim.core.utils import _chunk_like, uses_dask

__all__ = [
    "cosine_of_solar_zenith_angle",
    "day_angle",
    "day_lengths",
    "distance_from_sun",
    "eccentricity_correction_factor",
    "extraterrestrial_solar_radiation",
    "gladstones_day_length_latitude_coefficient",
    "huglin_day_length_latitude_coefficient",
    "jones_day_length_latitude_coefficient",
    "make_hourly_temperature",
    "resample_map",
    "solar_declination",
    "time_correction_for_solar_angle",
    "wind_speed_height_conversion",
]


def _wrap_radians(da):
    with xr.set_options(keep_attrs=True):
        return ((da + np.pi) % (2 * np.pi)) - np.pi


def distance_from_sun(dates: xr.DataArray) -> xr.DataArray:
    """
    Sun-earth distance.

    The distance from sun to earth in astronomical units.

    Parameters
    ----------
    dates : xr.DataArray
        Series of dates and time of days.

    Returns
    -------
    xr.DataArray, [astronomical units]
        Sun-earth distance.

    References
    ----------
    # TODO: Find a way to reference this
    U.S. Naval Observatory:Astronomical Almanac. Washington, D.C.: U.S. Government Printing Office (1985).
    """
    cal = get_calendar(dates)
    if cal == "default":
        cal = "standard"
    days_since = cftime.date2num(ensure_cftime_array(dates), "days since 2000-01-01 12:00:00", calendar=cal)
    g = ((357.528 + 0.9856003 * days_since) % 360) * np.pi / 180
    sun_earth = 1.00014 - 0.01671 * np.cos(g) - 0.00014 * np.cos(2.0 * g)
    return xr.DataArray(sun_earth, coords=dates.coords, dims=dates.dims)


def day_angle(time: xr.DataArray) -> xr.DataArray:
    """
    Day of year as an angle.

    Assuming the Earth makes a full circle in a year, this is the angle covered from
    the beginning of the year up to that timestep. Also called the "julian day fraction".

    Parameters
    ----------
    time : xr.DataArray
        Time coordinate.

    Returns
    -------
    xr.DataArray, [rad]
        Day angle.
    """
    if XR2409:
        decimal_year = time.dt.decimal_year
    else:
        decimal_year = _datetime_to_decimal_year(times=time, calendar=time.dt.calendar)
    return ((decimal_year % 1) * 2 * np.pi).assign_attrs(units="rad")


def solar_declination(time: xr.DataArray, method="spencer") -> xr.DataArray:
    """
    Solar declination.

    The angle between the sun rays and the earth's equator, in radians, as approximated
    by :cite:t:`spencer_fourier_1971` or assuming the orbit is a circle.

    Parameters
    ----------
    time : xr.DataArray
        Time coordinate.
    method : {'spencer', 'simple'}
        Which approximation to use. The default ("spencer") uses the first seven (7) terms of the
        Fourier series representing the observed declination, while "simple" assumes the orbit is
        a circle with a fixed obliquity and that the solstice/equinox happen at fixed angles on
        the orbit (the exact calendar date changes for leap years).

    Returns
    -------
    xr.DataArray, [rad]
        Solar declination angle.

    References
    ----------
    :cite:cts:`spencer_fourier_1971`
    """
    # julian day fraction
    da = convert_units_to(day_angle(time), "rad")
    if method == "simple":
        # This assumes the orbit is a perfect circle, the obliquity is 0.4091 rad (23.43°)
        # and the equinox is on the March 21st 17:20 UTC (March 20th 23:14 UTC on leap years)
        sd = 0.4091 * np.sin(da - 1.39)
    elif method == "spencer":
        sd = (
            0.006918
            - 0.399912 * np.cos(da)
            + 0.070257 * np.sin(da)
            - 0.006758 * np.cos(2 * da)
            + 0.000907 * np.sin(2 * da)
            - 0.002697 * np.cos(3 * da)
            + 0.001480 * np.sin(3 * da)
        )
    else:
        raise NotImplementedError("Method must be one of 'simple' or 'spencer'.")
    return _wrap_radians(sd).assign_attrs(units="rad").rename("declination")


def time_correction_for_solar_angle(time: xr.DataArray) -> xr.DataArray:
    """
    Time correction for solar angle.

    Every 1° of angular rotation on earth is equal to 4 minutes of time.
    The time correction is needed to adjust local watch time to solar time.

    Parameters
    ----------
    time : xr.DataArray
        Time coordinate.

    Returns
    -------
    xr.DataArray, [rad]
        Time correction of solar angle.

    References
    ----------
    :cite:cts:`di_napoli_mean_2020`
    """
    da = convert_units_to(day_angle(time), "rad")
    tc = (
        0.004297 + 0.107029 * np.cos(da) - 1.837877 * np.sin(da) - 0.837378 * np.cos(2 * da) - 2.340475 * np.sin(2 * da)
    )
    tc = tc.assign_attrs(units="degrees")
    return _wrap_radians(convert_units_to(tc, "rad"))


def eccentricity_correction_factor(
    time: xr.DataArray, method: Literal["spencer", "simple"] = "spencer"
) -> xr.DataArray:
    """
    Eccentricity correction factor of the Earth's orbit.

    The squared ratio of the mean distance Earth-Sun to the distance at a specific moment.
    As approximated by :cite:t:`spencer_fourier_1971`.

    Parameters
    ----------
    time : xr.DataArray
        Time coordinate.
    method : {'spencer', 'simple'}
        Which approximation to use.
        The default ("spencer") uses the first five (5) terms of the fourier series of the eccentricity.
        The "simple" method approximates with only the first two (2).

    Returns
    -------
    xr.DataArray, [dimensionless]
        Eccentricity correction factor.

    References
    ----------
    :cite:cts:`spencer_fourier_1971,perrin_estimation_1975`
    """
    # julian day fraction
    da = convert_units_to(day_angle(time), "rad")
    if method == "simple":
        # It is quite used, I think the source is (not available online):
        # Perrin de Brichambaut, C. (1975).
        # Estimation des ressources énergétiques solaires en France. Ed. Européennes thermique et industrie.
        return cast(xr.DataArray, 1 + 0.033 * np.cos(da))
    if method == "spencer":
        return cast(
            xr.DataArray,
            1.0001100
            + 0.034221 * np.cos(da)
            + 0.001280 * np.sin(da)
            + 0.000719 * np.cos(2 * da)
            + 0.000077 * np.sin(2 * da),
        )
    raise NotImplementedError("Method must be one of 'simple' or 'spencer'.")


def cosine_of_solar_zenith_angle(
    time: xr.DataArray,
    declination: xr.DataArray,
    lat: Quantified | xr.DataTree,
    lon: Quantified = "0 °",
    time_correction: xr.DataArray | None = None,
    stat: Literal["average", "integral", "instant"] = "average",
    sunlit: bool = False,
    chunks: dict[str, int] | None = None,
) -> xr.DataArray:
    """
    Cosine of the solar zenith angle.

    The solar zenith angle is the angle between a vertical line (perpendicular to the ground) and the sun rays.
    This function computes a statistic of its cosine : its instantaneous value,
    the integral from sunrise to sunset or the average over the same period or over a subdaily interval.
    Based on :cite:t:`kalogirou_chapter_2014` and :cite:t:`di_napoli_mean_2020`.

    Parameters
    ----------
    time : xr.DataArray
        The UTC time. If not daily and `stat` is "integral" or "average",
        the timestamp is taken as the start of interval.
        If daily, the interval is assumed to be centered on Noon.
        If fewer than three timesteps are given, a daily frequency is assumed.
    declination : xr.DataArray
        Solar declination. See :py:func:`solar_declination`.
    lat : Quantified
        Latitude coordinate. Expects units of "degree_north".
    lon : Quantified
        Longitude. Needed if the input timeseries is subdaily.
    time_correction : xr.DataArray, optional
        Time correction for solar angle. See :py:func:`time_correction_for_solar_angle`
        This is necessary if stat is "instant".
    stat : {'average', 'integral', 'instant'}
        Which daily statistic to return.
        If "average", this returns the average of the cosine of the zenith angle
        If "integral", this returns the integral of the cosine of the zenith angle
        If "instant", this returns the instantaneous cosine of the zenith angle
    sunlit : bool
        If True, only the sunlit part of the interval is considered in the integral or average.
        Does nothing if stat is "instant".
    chunks : dictionary
        When `time`,  `lat` and `lon` originate from coordinates of a large chunked dataset, this dataset's chunking
        can be passed here to ensure the computation is also chunked.

    Returns
    -------
    xr.DataArray, [rad] or [dimensionless]
        Cosine of the solar zenith angle. If stat is "integral", dimensions can be said to be "time" as the integral
        is on the hour angle.
        For seconds, multiply by the number of seconds in a complete day cycle (24*60*60) and divide by 2π.

    Notes
    -----
    This code was inspired by the `thermofeel` and `PyWBGT` package.

    References
    ----------
    :cite:cts:`kalogirou_chapter_2014,di_napoli_mean_2020`
    """
    declination = convert_units_to(declination, "rad")
    lat = _wrap_radians(convert_units_to(lat, "rad"))
    lon = convert_units_to(lon, "rad")
    declination, lat, lon = _chunk_like(declination, lat, lon, chunks=chunks)

    S_IN_D = 24 * 3600

    if len(time) < 3 or xr.infer_freq(time) == "D":
        h_s = -np.pi if stat != "instant" else 0
        h_e = np.pi - 1e-9  # just below pi
    else:
        if time.dtype == "O":  # cftime
            time_as_s = time.copy(data=xr.CFTimeIndex(cast(CFTimeIndex, time.values)).asi8 / 1e6)
        else:  # numpy
            time_as_s = time.copy(data=time.astype(float) / 1e9)
        h_s_utc = (((time_as_s % S_IN_D) / S_IN_D) * 2 * np.pi + np.pi).assign_attrs(units="rad")
        h_s = h_s_utc + lon

        interval_as_s = time.diff("time").dt.seconds.reindex(time=time.time, method="bfill")
        h_e = h_s + 2 * np.pi * interval_as_s / S_IN_D

    if stat == "instant":
        h_s = h_s + time_correction

        return cast(
            xr.DataArray,
            np.sin(declination) * np.sin(lat) + np.cos(declination) * np.cos(lat) * np.cos(h_s),
        ).clip(0, None)
    if stat not in {"average", "integral"}:
        raise NotImplementedError("Argument 'stat' must be one of 'integral', 'average' or 'instant'.")
    if sunlit:
        # hour angle of sunset (eq. 2.15), with NaNs inside the polar day/night
        tantan = cast(xr.DataArray, -np.tan(lat) * np.tan(declination))
        h_ss = np.arccos(tantan.where(abs(tantan) <= 1))
    else:
        # Whole period, so we put sunset at midnight
        h_ss = np.pi - 1e-9

    return xr.apply_ufunc(
        _sunlit_integral_of_cosine_of_solar_zenith_angle,
        declination,
        lat,
        _wrap_radians(h_ss),
        _wrap_radians(h_s),
        _wrap_radians(h_e),
        stat == "average",
        input_core_dims=[[]] * 6,
        dask="parallelized",
    )


@nb.vectorize
def _sunlit_integral_of_cosine_of_solar_zenith_angle(declination, lat, h_sunset, h_start, h_end, average):
    """Integral of the cosine of the solar zenith angle over the sunlit part of the interval."""
    # Code inspired by PyWBGT
    h_sunrise = -h_sunset
    # Polar day
    if np.isnan(h_sunset) & ((declination * lat) > 0):
        num = np.sin(h_end) - np.sin(h_start)
        # Polar day with interval crossing midnight
        if h_end < h_start:
            denum = h_end + 2 * np.pi - h_start
        else:
            denum = h_end - h_start
    # Polar night:
    elif np.isnan(h_sunset) & ((declination * lat) < 0):
        return 0
    # No sunlit interval (at night) 1) crossing midnight and 2) between 0h and sunrise 3) between sunset and 0h
    elif (
        (h_start > h_sunset and h_end < h_sunrise)
        or (h_start < h_sunrise and h_end < h_sunrise)
        or (h_start > h_sunset and h_end > h_sunset)
    ):
        return 0
    # Interval crossing midnight, starting after sunset (before midnight), finishing after sunrise
    elif h_start > h_end >= h_sunrise and h_start >= h_sunset:
        num = np.sin(h_end) - np.sin(h_sunrise)
        denum = h_end - h_sunrise
    # Interval crossing midnight, starting after sunrise, finishing after sunset (after midnight)
    elif h_end < h_start and h_start >= h_sunrise >= h_end:
        num = np.sin(h_sunset) - np.sin(h_start)
        denum = h_sunset - h_start
    # Interval crossing midnight, starting before sunset, finishing after sunrise (2 sunlit parts)
    elif h_sunset >= h_start > h_end >= h_sunrise:
        num = np.sin(h_sunset) - np.sin(h_start) + np.sin(h_end) - np.sin(h_sunrise)
        denum = h_sunset - h_start + h_end - h_sunrise
    # All other cases : interval not crossing midnight, overlapping with the sunlit part
    else:
        h1 = max(h_sunrise, h_start)
        h2 = min(h_sunset, h_end)
        num = np.sin(h2) - np.sin(h1)
        denum = h2 - h1
    out = np.sin(declination) * np.sin(lat) * denum + np.cos(declination) * np.cos(lat) * num
    if average:
        out = out / denum
    return out


def extraterrestrial_solar_radiation(
    times: xr.DataArray,
    lat: xr.DataArray,
    solar_constant: Quantified = "1361 W m-2",
    method: Literal["spencer", "simple"] = "spencer",
    chunks: Mapping[Any, tuple] | None = None,
) -> xr.DataArray:
    """
    Extraterrestrial solar radiation.

    This is the daily energy received on a surface parallel to the ground at the mean distance of the earth to the sun.
    It neglects the effect of the atmosphere. Computation is based on :cite:t:`kalogirou_chapter_2014` and the default
    solar constant is taken from :cite:t:`matthes_solar_2017`.

    Parameters
    ----------
    times : xr.DataArray
        Daily datetime data. This function makes no sense with data of other frequency.
    lat : xr.DataArray
        Latitude coordinate. Expects units of "degree_north".
    solar_constant : str
        The solar constant, the energy received on earth from the sun per surface per time.
    method : {'spencer', 'simple'}
        Which method to use when computing the solar declination and the eccentricity correction factor.
        See :py:func:`solar_declination` and :py:func:`eccentricity_correction_factor`.
    chunks : dict
        When `times` and `lat` originate from coordinates of a large chunked dataset, passing the dataset's chunks here
        will ensure the computation is chunked as well.

    Returns
    -------
    xr.DataArray, [J m-2 d-1]
        Extraterrestrial solar radiation.

    References
    ----------
    :cite:cts:`kalogirou_chapter_2014,matthes_solar_2017`
    """
    dr = eccentricity_correction_factor(times, method=method)
    ds = solar_declination(times, method=method)
    gsc = convert_units_to(solar_constant, "J m-2 d-1")
    rad_to_day = 1 / (2 * np.pi)  # convert radians of the "day circle" to day
    return (
        gsc
        * rad_to_day
        * cosine_of_solar_zenith_angle(times, ds, lat=lat, stat="integral", sunlit=True, chunks=chunks)
        * dr
    ).assign_attrs(units="J m-2 d-1")


def day_lengths(
    dates: xr.DataArray,
    lat: Quantified | xr.Dataset | xr.DataTree,
    method: Literal["spencer", "simple"] = "spencer",
    infill_polar_days: bool = False,
) -> xr.DataArray:
    r"""
    Calculate day-length according to latitude and day of the year.

    See :py:func:`solar_declination` for the approximation used to compute the solar declination angle.
    Based on :cite:t:`kalogirou_chapter_2014`.

    Parameters
    ----------
    dates : xr.DataArray
        Daily datetime data.
        This function makes no sense with data of other time frequencies.
    lat : Quantified or xarray.Dataset or xarray.DataTree
        Latitude coordinate. Expects units of "degree_north".
    method : {'spencer', 'simple'}
        Which approximation to use when computing the solar declination angle.
        See :py:func:`xclim.indices.helpers.solar_declination`.
    infill_polar_days : bool
        Whether to use a mask of 24 hours for polar days and 0 hours for polar nights.
        If False, polar days and nights will be NaN.
        If True, they will be filled with 24 and 0 hours, respectively,
        dependent on latitude and solar declination at the given date.

    Returns
    -------
    xarray.DataArray, [hours]
        Day-lengths in hours per individual day.

    Raises
    ------
    NotImplementedError
        If a series of dates provided are not inferrable at a daily time frequency.

    Notes
    -----
    The day length is computed as the time between sunrise and sunset.
    The infill_polar_days option provides an arbitrary method fill polar days and nights with
    24 and 0 hours, respectively. Care should be taken when using this option, as it may not be
    appropriate for all applications.

    References
    ----------
    :cite:cts:`kalogirou_chapter_2014`
    """
    if len(dates) >= 3 and xr.infer_freq(dates) != "D":
        raise NotImplementedError("day_lengths only supports daily data.")

    declination = solar_declination(dates.time, method=method)
    radians = convert_units_to(lat, "rad")
    lat_deg = convert_units_to(lat, "deg")
    # arccos gives the hour-angle at sunset, multiply by 24 / 2π to get hours.
    # The day length is twice that.
    with np.errstate(invalid="ignore"):
        day_length_hours = ((24 / np.pi) * np.arccos(-np.tan(radians) * np.tan(declination))).assign_attrs(units="h")

    if infill_polar_days:
        lat_broadcast, decl_broadcast = xr.broadcast(lat_deg, declination)
        # Polar day: sun never sets; Polar night: sun never rises.
        polar_day = ((lat_broadcast > 66.5) & (decl_broadcast > 0)) | ((lat_broadcast < -66.5) & (decl_broadcast < 0))
        polar_night = ((lat_broadcast > 66.5) & (decl_broadcast < 0)) | ((lat_broadcast < -66.5) & (decl_broadcast > 0))
        # Infill polar days with 24 hours and polar nights with 0 hours.
        valid = ~xr.ufuncs.isnan(day_length_hours)
        day_length_hours = day_length_hours.where(~(polar_day & ~valid), 24.0)
        day_length_hours = day_length_hours.where(~(polar_night & ~valid), 0)

    # Drop nonessential coordinates
    for coord in day_length_hours.coords:
        if coord not in ["lat", "time"]:
            day_length_hours = day_length_hours.drop_vars(coord)

    return day_length_hours


def huglin_day_length_latitude_coefficient(
    lat: xr.DataArray | str,
    method: Literal["huglin", "interpolated"],
    cap_value: float = np.nan,
) -> xr.DataArray:
    r"""
    Simple coefficient for the day-length and high latitudes.

    This latitude coefficient is used for determining the latitude effect on the day length specific to climate indices
    that concern viticulture, such as :py:func:`xclim.indices.huglin_index` (cite:p:`huglin_nouveau_1978`).
    This function is an empirical approximation of the day-length multiplication factor, :math:`k`, based on latitude.

    Parameters
    ----------
    lat : xarray.DataArray, str
        Latitude coordinate. Expects units of "degree_north".
        If provided a string (e.g. "45 degree_north"), it is converted to an xarray.DataArray.
    method : {"huglin", "interpolated"}
        The method to use for the coefficient calculation.
    cap_value : float
        For latitudes north of 50° N and south of 50° S, the value for the coefficient.

    Returns
    -------
    xarray.DataArray, [dimensionless]
        Coefficient for the day length based on latitude.

    Notes
    -----
    For the original `"huglin"` implementation :cite:p:`huglin_nouveau_1978`, the day-length multiplication factor,
    :math:`k`, is calculated as follows:

    .. math::

       k = f(lat) = \begin{cases}
                     1.0, & \text{if } | lat | <= 40 \\
                     1.02, & \text{if } 40 < | lat | <= 42 \\
                     1.03, & \text{if } 42 < | lat | <= 44 \\
                     1.04, & \text{if } 44 < | lat | <= 46 \\
                     1.05, & \text{if } 46 < | lat | <= 48 \\
                     1.06, & \text{if } 48 < | lat | <= 50 \\
                     m, & \text{if } | lat | > 50 \\
                     \end{cases}

    An alternative implementation (`"interpolated"`) uses smoothing to reduce the stepwise behaviour of the
    "huglin"` method. The day-length multiplication factor (:math:`k`) for the `"interpolated"` method then is
    calculated as follows:

    .. math::

       k = f(lat) = \begin{cases}
                     1, & \text{if } | lat | <= 40 \\
                     1 + ((abs(lat) - 40) / 10) * 0.06, & \text{if } 40 < | lat | <= 50 \\
                     m, & \text{if } | lat | > 50 \\
                     \end{cases}

    Where :math:`m` is the cap value, which is set to np.nan, or other if provided.

    References
    ----------
    :cite:cts:`huglin_nouveau_1978,project_team_eca&d_algorithm_2013`
    """
    if isinstance(lat, str):
        _lat_value = convert_units_to(lat, "deg")
        _lat = xr.DataArray(lat, attrs={"units": "degree_north"})
    else:
        _lat = lat

    if isinstance(cap_value, float):
        _cap_value = cap_value
    else:
        raise TypeError("Argument 'cap_value' must be a float (or numpy.nan).")

    lat_abs = abs(lat)
    if method == "huglin":
        k_f_bounds = [(0, -np.inf, 40), (0.02, 40, 42), (0.03, 42, 44), (0.04, 44, 46), (0.05, 46, 48), (0.06, 48, 50)]
        k = xr.full_like(lat_abs, _cap_value + 1)
        for k_f_b in k_f_bounds:
            cond = (k_f_b[1] < lat_abs) & (lat_abs <= k_f_b[2])
            k = k.where(~cond, 1 + k_f_b[0])
    elif method == "interpolated":
        lat_mask = lat_abs <= 50
        lat_coefficient = 1 + ((lat_abs - 40) / 10).clip(min=0) * 0.06
        k = xr.where(lat_mask, lat_coefficient, _cap_value)
    else:
        raise NotImplementedError("Method is not implemented. Only 'huglin' and 'interpolated' are permitted.")

    return k


def gladstones_day_length_latitude_coefficient(
    dates: xr.DataArray,
    lat: xr.DataArray | int | float,
    neutral_latitude: str = "40.0 deg",
    constrain: str | None = None,
    day_length_method: Literal["simple", "spencer"] = "spencer",
) -> xr.DataArray:
    """
    Day-length latitude coefficient based on the Gladstones methodology.

    This function computes a day-length latitude coefficient as it influences the monthly temperatures
    of the growing season as compared to the day-length of a neutral reference latitude.
    Based on :cite:t:`gladstones_viticulture_1992` and cite:t:`gladstones_wine_2011`.

    Parameters
    ----------
    dates : xarray.DataArray
        The dates for which the day length latitude coefficient is computed.
    lat : xarray.DataArray or int or float
        Latitude coordinate. Expects units of "degree_north".
        If a single value is given, it is converted to an xarray.DataArray.
    neutral_latitude : str
        The latitude at which the day length coefficient is 1.0.
        Latitudes between this value and 0 degrees North will have a coefficient below 1.0 during the growing season,
        while latitudes above this value will have a coefficient greater than 1.0.
        This negative absolute value of this latitude is used for calculating coefficients in the Southern Hemisphere.
    constrain : str, optional
        The lower latitude limit for applying the latitude coefficient.
        If a str is given (e.g. '25 degree_north`), values below this threshold will be set to '1.0'.
    day_length_method : {'simple', 'spencer'}
        The method to use for the day length calculation.
        The "simple" method uses a simple approximation of the day length based on latitude and time of year.
        The "spencer" method uses a more complex approximation based on the Fourier series of the solar declination.

    Returns
    -------
    xarray.DataArray, [dimensionless]
        Coefficient for the day length based on latitude.
    """
    if not isinstance(lat, xr.DataArray):
        lat = xr.DataArray(lat, attrs={"units": "degree_north"})
    if day_length_method not in ["simple", "spencer"]:
        raise NotImplementedError("day_length_method must be one of 'simple' or 'spencer'.")

    if isinstance(constrain, str):
        constrain_value = convert_units_to(constrain, "deg")
    elif constrain is None:
        constrain_value = False
    else:
        raise ValueError("Argument 'constrain' must be a str (e.g. '25 degree_north') or 'None'.")

    _neutral_latitude = convert_units_to(neutral_latitude, "deg")

    pivotal_day_length_north = day_lengths(dates=dates, lat=f"{abs(_neutral_latitude)} deg", method=day_length_method)
    pivotal_day_length_south = day_lengths(dates=dates, lat=f"{-abs(_neutral_latitude)} deg", method=day_length_method)
    day_length = day_lengths(dates=dates, lat=lat, method=day_length_method)

    if not constrain_value:
        k = xr.where(lat >= 0.0, day_length / pivotal_day_length_north, day_length / pivotal_day_length_south)
    else:
        k = xr.where(
            lat >= constrain_value,
            day_length / pivotal_day_length_north,
            xr.where(lat <= -constrain_value, day_length / pivotal_day_length_south, 1.0),
        )
    k = k.assign_attrs(units="dimensionless")

    return k


def jones_day_length_latitude_coefficient(
    dates: xr.DataArray,
    lat: xr.DataArray | xr.Dataset | xr.DataTree,
    method: Literal["gladstones", "jones"],
    floor: bool = False,
    start_date: str | DayOfYearStr = "04-01",
    end_date: str | DayOfYearStr = "11-01",
    freq: Literal["YS", "YS-JAN", "YS-JUL"] = "YS",
) -> xr.DataArray:
    r"""
    Complex day length latitude coefficient.

    This function computes a day length latitude coefficient as it influences the entire growing season.
    Based on cite:t:`hall_spatial_2010`.

    Parameters
    ----------
    dates : xarray.DataArray
        The dates for which the day length latitude coefficient is computed.
    lat : xr.DataArray or xr.Dataset or xr.DataTree
        Latitude coordinate. Expects units of "degree_north".
        If a single value is given, it is converted to an xarray.DataArray.
    method : {"gladstones", "jones"}
        The method to use for the coefficient calculation.
        The "jones" method .
        The "gladstones" method uses an approximation of the Gladstones methodology for day length latitude coefficient.
    floor : bool, optional
        If True, latitudes where the day length latitude coefficient would be below '1.0', the value is set to '1.0'.
        if False, coefficient can be below '1.0' for latitudes where the day length is less than the reference latitude.
    start_date : str or DayOfYearStr
        The start date of the growing season.
    end_date : str or DayOfYearStr
        The end date of the growing season. Date is not included in the aggregation.
    freq : {"YS", "YS-JAN", "YS-JUL"}
        The frequency at which to aggregate the day lengths.
        Must be an annual frequency, such as "YS" or "YS-JAN" (yearly start in January)
        or "YS-JUL" (yearly start in July).

    Returns
    -------
    xarray.DataArray, [dimensionless]
        Coefficient for the day length based on latitude, aggregated over the growing season.

    Raises
    ------
    ValueError
        If all latitudes for every computed growing season have a day length latitude coefficient below 1.0.

    Notes
    -----
    For the `"jones"` method, A more robust day-length calculation based on latitude, calendar, day-of-year, and
    obliquity is used. This algorithm requires a calculation of the sum of the day lengths over the growing season
    at each latitude, :math:`totalSeasonDayLength_{Lat}`, which is then used to calculate the day length latitude
    coefficient :math:`k`:

    .. math::

        totalSeasonDayLength_{Lat} = \sum_{Jday=\text{103}}^{\text{284}}{dayLength_{Lat_{JDay}}}

    The day length latitude coefficient (:math:`k`) using the "jones" method is calculated as follows:

    .. math::

         k_{Lat} = 2.8311e-4 * totalSeasonDayLength_{Lat} + 0.30834

    The "gladstones" method provided here is a transformation of the "jones" method, based on the relationship
    detailed in :cite:t:`hall_spatial_2010`:

    .. math::

        k_{Lat}^{Gladstones} = 1.1135 * k_{Lat}^{Jones} - 0.1352

    For both of these methods, the :math:`k` coefficient must be calculated at the growing season frequency (yearly),
    starting from either January or July, depending on the hemisphere of interest.
    """
    if parse_offset(freq) not in [(1, "Y", True, "JAN"), (1, "Y", True, "JUL")]:
        msg = (
            f"Freq {freq} not supported. Must be 'YS'/'YS-JAN', or 'YS-JUL' for method 'jones'. "
            "An annual frequency is required for the current implementation."
        )
        raise NotImplementedError(msg)

    if method in ["gladstones", "jones"]:
        if freq == "YS-JUL":
            pass
        day_length = (
            select_time(
                day_lengths(dates=dates, lat=lat, method="spencer", infill_polar_days=False),
                date_bounds=(start_date, end_date),
                include_bounds=(True, False),
            )
            .dropna(dim="time", how="all")
            .dropna(dim="lat", how="any")
            .resample(time=freq)
            .sum()
        )
        k_jones: xr.DataArray = (2.8311e-4 * day_length) + 0.30834

        all_below_1 = (k_jones < 1.0).all(dim="lat")
        k_jones = k_jones.where(~all_below_1, np.nan)
        if k_jones.isnull().all():
            msg = (
                "All latitudes for every growing season have a day length latitude coefficient below 1.0. "
                "This is likely due to the start and end dates of the growing season being too restrictive "
                "or an incomplete time series."
            )
            raise ValueError(msg)

        if method == "jones":
            k_aggregated = k_jones
        else:
            k_aggregated = 1.1135 * k_jones - 0.1352
    else:
        raise NotImplementedError("Method not implemented. Only 'gladstones' or 'jones' are supported.")

    if floor:
        k_aggregated = k_aggregated.where(k_aggregated >= 1.0, 1.0)

    return k_aggregated


def wind_speed_height_conversion(
    ua: xr.DataArray,
    h_source: str,
    h_target: str,
    method: Literal["log"] = "log",
) -> xr.DataArray:
    r"""
    Wind speed at two meters.

    Parameters
    ----------
    ua : xarray.DataArray
        Wind speed at height `h`.
    h_source : str
        Height of the input wind speed `ua` (e.g. `h == "10 m"` for a wind speed at `10 meters`).
    h_target : str
        Height of the output wind speed.
    method : {"log"}
        Method used to convert wind speed from one height to another.

    Returns
    -------
    xarray.DataArray
        Wind speed at height `h_target`.

    References
    ----------
    :cite:cts:`allen_crop_1998`
    """
    h_source = convert_units_to(h_source, "m")
    h_target = convert_units_to(h_target, "m")
    if method == "log":
        if min(h_source, h_target) < 1 + 5.42 / 67.8:
            raise ValueError(
                f"The height {min(h_source, h_target)}m is too small for method {method}. "
                f"Heights must be greater than {1 + 5.42 / 67.8}"
            )
        with xr.set_options(keep_attrs=True):
            return ua * np.log(67.8 * h_target - 5.42) / np.log(67.8 * h_source - 5.42)
    else:
        raise NotImplementedError(f"'{method}' method is not implemented.")


def _gather_lat(da: xr.DataArray) -> xr.DataArray:
    """
    Gather latitude coordinate using cf-xarray.

    Parameters
    ----------
    da : xarray.DataArray
        CF-conformant DataArray with a "latitude" coordinate.

    Returns
    -------
    xarray.DataArray
        Latitude coordinate.
    """
    try:
        lat = da.cf["latitude"]
        return lat
    except KeyError as err:
        n_func = stack()[1].function
        msg = f"{n_func} could not find latitude coordinate in DataArray. Try passing it explicitly (`lat=ds.lat`)."
        raise ValueError(msg) from err


def _gather_lon(da: xr.DataArray) -> xr.DataArray:
    """
    Gather longitude coordinate using cf-xarray.

    Parameters
    ----------
    da : xarray.DataArray
        CF-conformant DataArray with a "longitude" coordinate.

    Returns
    -------
    xarray.DataArray
        Longitude coordinate.
    """
    try:
        lat = da.cf["longitude"]
        return lat
    except KeyError as err:
        n_func = stack()[1].function
        msg = f"{n_func} could not find longitude coordinate in DataArray. Try passing it explicitly (`lon=ds.lon`)."
        raise ValueError(msg) from err


def resample_map(
    obj: xr.DataArray | xr.Dataset,
    dim: str,
    freq: str,
    func: Callable,
    map_blocks: bool | Literal["from_context"] = "from_context",
    resample_kwargs: dict | None = None,
    map_kwargs: dict | None = None,
) -> xr.DataArray | xr.Dataset:
    r"""
    Wrap xarray's resample(...).map() with a :py:func:`xarray.map_blocks`.

    Ensures that the chunking is appropriate using `flox`.

    Parameters
    ----------
    obj : DataArray or Dataset
        The xarray object to resample.
    dim : str
        Dimension over which to resample.
    freq : str
        Resampling frequency along `dim`.
    func : callable
        Function to map on each resampled group.
    map_blocks : bool or "from_context"
        If True, the resample().map() call is wrapped inside a `map_blocks`.
        If False, this does not do anything special.
        If "from_context", xclim's "resample_map_blocks" option is used.
        If the object is not using dask, this is set to False.
    resample_kwargs : dict, optional
        Other arguments to pass to `obj.resample()`.
    map_kwargs : dict, optional
        Arguments to pass to `map`.

    Returns
    -------
    xr.DataArray or xr.Dataset
        Resampled object.
    """
    resample_kwargs = resample_kwargs or {}
    map_kwargs = map_kwargs or {}
    if map_blocks == "from_context":
        map_blocks = OPTIONS[MAP_BLOCKS]

    if not uses_dask(obj) or not map_blocks:
        return obj.resample({dim: freq}, **resample_kwargs).map(func, **map_kwargs)

    if rechunk_for_blockwise is None:
        msg = f"Using {MAP_BLOCKS}=True requires flox."
        raise ValueError(msg) from flox_err

    # Make labels, a unique integer for each resample group
    labels = xr.full_like(obj[dim], -1, dtype=np.int32)
    for lbl, group_slice in enumerate(obj[dim].resample({dim: freq}).groups.values()):
        labels[group_slice] = lbl

    obj_rechunked = rechunk_for_blockwise(obj, dim, labels)

    def _resample_map(obj_chnk, dm, frq, rs_kws, fun, mp_kws):
        return obj_chnk.resample({dm: frq}, **rs_kws).map(fun, **mp_kws)

    # Template. We are hoping that this takes a negligeable time as it is never loaded.
    template = obj_rechunked.resample(**{dim: freq}, **resample_kwargs).first()

    # New chunks along the time dim : infer the number of elements resulting from the resampling of each chunk
    if isinstance(obj_rechunked, xr.Dataset):
        chunksizes = obj_rechunked.chunks[dim]
    else:
        chunksizes = obj_rechunked.chunks[obj_rechunked.get_axis_num(dim)]
    new_chunks = []
    i = 0
    for chunksize in chunksizes:
        new_chunks.append(len(np.unique(labels[i : i + chunksize])))
        i += chunksize
    template = template.chunk({dim: tuple(new_chunks)})

    return obj_rechunked.map_blocks(_resample_map, (dim, freq, resample_kwargs, func, map_kwargs), template=template)


def _compute_daytime_temperature(
    hour_after_sunrise: xr.DataArray,
    tasmin: xr.DataArray,
    tasmax: xr.DataArray,
    daylength: xr.DataArray,
) -> xr.DataArray:
    """
    Compute daytime temperature based on a sinusoidal profile.

    Minimum temperature is reached at sunrise and maximum temperature 2h before sunset.

    Parameters
    ----------
    hour_after_sunrise : xarray.DataArray
        Hours after the last sunrise.
    tasmin : xarray.DataArray
        Daily minimum temperature.
    tasmax : xarray.DataArray
        Daily maximum temperature.
    daylength : xarray.DataArray
        Length of the day in hours.

    Returns
    -------
    xarray.DataArray
        Hourly daytime temperature.
    """
    return (tasmax - tasmin) * np.sin((np.pi * hour_after_sunrise) / (daylength + 4)) + tasmin


def _compute_nighttime_temperature(
    hours_after_sunset: xr.DataArray,
    tasmin: xr.DataArray,
    tas_sunset: xr.DataArray,
    daylength: xr.DataArray,
) -> xr.DataArray:
    """
    Compute nighttime temperature based on a logarithmic profile.

    Temperature at sunset is computed from previous daytime temperature,
    minimum temperature is reached at sunrise.

    Parameters
    ----------
    hours_after_sunset : xarray.DataArray
        Hours after the last sunset.
    tasmin : xarray.DataArray
        Daily minimum temperature.
    tas_sunset : xarray.DataArray
        Temperature at last sunset.
    daylength : xarray.DataArray
        Length of the day in hours.

    Returns
    -------
    xarray.DataArray
        Hourly nighttime temperature.
    """
    return tas_sunset - ((tas_sunset - tasmin) / np.log(24 - daylength)) * np.log(hours_after_sunset)


def _add_one_day(time: xr.DataArray) -> xr.DataArray:
    """
    Add one day to a time coordinate.

    Depending on the calendar/dtype of the time array we need to use numpy's or datetime's (for cftimes) timedelta.

    Parameters
    ----------
    time : xr.DataArray
        Time coordinate.

    Returns
    -------
    xr.DataArray
        Next day.
    """
    if time.dtype == "O":
        return time + timedelta(days=1)
    return time + np.timedelta64(1, "D")


def make_hourly_temperature(tasmin: xr.DataArray, tasmax: xr.DataArray) -> xr.DataArray:
    """
    Compute hourly temperatures from tasmin and tasmax.

    Based on the Linvill et al. "Calculating Chilling Hours and Chill Units from Daily
    Maximum and Minimum Temperature Observations", HortScience, 1990
    we assume a sinusoidal temperature profile during daylight and a logarithmic decrease after sunset
    with tasmin reached at sunsrise and tasmax reached 2h before sunset.

    For simplicity and because it's used for daily aggregation, we assume that sunrise globally happens at midnight
    and the sunsets after `daylength` hours computed via the :py:func:`day_lengths` function.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Daily minimum temperature.
    tasmax : xarray.DataArray
        Daily maximum temperature.

    Returns
    -------
    xarray.DataArray
        Hourly temperature.
    """
    data = xr.merge([tasmin, tasmax])
    data = data.assign_coords(time=data.time.dt.floor("D"))
    # We add one more timestamp so the resample function includes the last day
    data = xr.concat(
        [
            data,
            data.isel(time=-1).assign_coords(time=_add_one_day(data.isel(time=-1).time)),
        ],
        dim="time",
    )

    daylength = day_lengths(data.time, data.lat)
    # Create daily chunks to avoid memory issues after the resampling
    data = data.assign(
        daylength=daylength,
        sunset_temp=_compute_daytime_temperature(daylength, data.tasmin, data.tasmax, daylength),
        next_tasmin=data.tasmin.shift(time=-1),
    )
    # Compute hourly data by resampling and remove the last time stamp that was added earlier
    hourly = data.resample(time="h").ffill().isel(time=slice(0, -1))

    # To avoid "invalid value encountered in log" warning we set hours before sunset to 1
    nighttime_hours = (hourly.time.dt.hour + 1 - hourly.daylength).clip(1)

    return xr.where(
        hourly.time.dt.hour < hourly.daylength,
        _compute_daytime_temperature(hourly.time.dt.hour, hourly.tasmin, hourly.tasmax, hourly.daylength),
        _compute_nighttime_temperature(
            nighttime_hours,
            hourly.next_tasmin,
            hourly.sunset_temp,
            hourly.daylength - 1,
        ),
    ).assign_attrs(units=tasmin.units)



from __future__ import annotations

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from xclim.core.options import set_options
from xclim.core.units import convert_units_to
from xclim.core.utils import uses_dask
from xclim.indices import helpers
from xclim.testing.helpers import assert_lazy


@pytest.mark.parametrize("method,rtol", [("spencer", 5e3), ("simple", 1e2)])
def test_solar_declinaton(method, rtol):
    # Expected values from https://gml.noaa.gov/grad/solcalc/azel.html
    times = xr.DataArray(
        pd.to_datetime(["1793-01-21T10:22:00", "1969-07-20T20:17:40", "2022-05-20T16:55:48"]),
        dims=("time",),
    )
    exp = [-19.83, 20.64, 20.00]
    np.testing.assert_allclose(
        helpers.solar_declination(times, method=method),
        np.deg2rad(exp),
        atol=rtol * 2 * np.deg2rad(23.44),  # % of the possible range
    )


@pytest.mark.parametrize("method", ["spencer", "simple"])
def test_extraterrestrial_radiation(method):
    # Expected values from https://www.engr.scu.edu/~emaurer/tools/calc_solar_cgi.pl
    # This source is not authoritative, thus the large rtol
    times = xr.DataArray(
        xr.date_range("1900-01-01", "1900-01-03", freq="D"),
        dims=("time",),
        name="time",
    )
    lat = xr.DataArray(
        [48.8656, 29.5519, -54],
        dims=("time",),
        coords={"time": times},
        attrs={"units": "degree_north"},
    )
    exp = [99.06, 239.98, 520.01]
    np.testing.assert_allclose(
        convert_units_to(helpers.extraterrestrial_solar_radiation(times, lat, method=method), "W m-2"),
        exp,
        rtol=3e-2,
    )


class TestDayLength:
    @staticmethod
    def data_setup(lats: np.ndarray, start_date: str = "1992-12-01", end_date: str = "1994-01-01"):
        time_data = xr.date_range(start_date, end_date, freq="D", calendar="standard")
        data = xr.DataArray(
            np.ones((time_data.size, len(lats))),
            dims=("time", "lat"),
            coords={"time": time_data, "lat": lats},
        )
        data.lat.attrs["units"] = "degree_north"
        return data

    @pytest.mark.parametrize("method, infill", [("spencer", False), ("spencer", True), ("simple", False)])
    def test_day_lengths(self, method, infill):
        data = self.data_setup(lats=np.array([-60, -45, -30, 0, 30, 45, 60, 80]))
        dl = helpers.day_lengths(dates=data.time, lat=data.lat, method=method, infill_polar_days=infill)

        events = dict(
            solstice=[
                ["1992-12-21", [18.49, 15.43, 13.93, 12.0, 10.07, 8.57, 5.51, 0.0 if infill else np.nan]],
                ["1993-06-21", [5.51, 8.57, 10.07, 12.0, 13.93, 15.43, 18.49, 24.0 if infill else np.nan]],
                ["1993-12-21", [18.49, 15.43, 13.93, 12.0, 10.07, 8.57, 5.51, 0.0 if infill else np.nan]],
            ],
            equinox=[
                ["1993-03-20", [12] * 8]
            ],  # True equinox on 1993-03-20 at 14:41 GMT. Some relative tolerance is needed.
        )

        for event, evaluations in events.items():
            for e in evaluations:
                if event == "solstice":
                    np.testing.assert_array_almost_equal(dl.sel(time=e[0]).transpose(), np.array(e[1]), 2)
                elif event == "equinox":
                    np.testing.assert_allclose(dl.sel(time=e[0]).transpose(), np.array(e[1]), rtol=2e-1)

    @pytest.mark.parametrize(
        "method, cap_value, results",
        [
            ("huglin", np.nan, [np.nan, 1.04, 1.03, 1.0, 1.03, 1.04, np.nan, np.nan]),
            ("interpolated", np.nan, [np.nan, 1.03, 1.02, 1.0, 1.02, 1.03, np.nan, np.nan]),
            ("interpolated", 1.06, [1.06, 1.03, 1.02, 1.0, 1.02, 1.03, 1.06, 1.06]),
        ],
    )
    def test_huglin_day_length_latitude_coefficient(self, method, cap_value, results):
        data = self.data_setup(lats=np.array([-60, -45, -43.5, 0, 43.5, 45, 60, 80]))
        k = helpers.huglin_day_length_latitude_coefficient(lat=data.lat, method=method, cap_value=cap_value)

        np.testing.assert_array_almost_equal(k, np.array(results), decimal=2)

    @pytest.mark.parametrize(
        "method, start_date, end_date, freq, floor, results",
        [
            (
                "gladstones",
                "04-01",
                "11-01",
                "YS",
                False,
                [0.75, 0.86, 0.91, 0.95, 0.97, 1.0, 1.02, 1.04, 1.06, 1.09, 1.12, 1.18, 1.29],
            ),
            (
                "gladstones",
                "04-01",
                "11-01",
                "YS-JAN",
                True,
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.02, 1.04, 1.06, 1.09, 1.12, 1.18, 1.29],
            ),
            (
                "gladstones",
                "10-01",
                "04-01",
                "YS-JUL",
                True,
                [1.18, 1.06, 1.01, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            ),
            (
                "jones",
                "04-01",
                "11-01",
                "YS-JAN",
                False,
                [0.79, 0.89, 0.94, 0.97, 1.0, 1.02, 1.04, 1.05, 1.07, 1.1, 1.13, 1.18, 1.28],
            ),
            (
                "jones",
                "04-01",
                "11-01",
                "YS",
                True,
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.02, 1.04, 1.05, 1.07, 1.1, 1.13, 1.18, 1.28],
            ),
            (
                "jones",
                "10-01",
                "04-01",
                "YS-JUL",
                False,
                [1.18, 1.07, 1.02, 0.99, 0.97, 0.95, 0.93, 0.91, 0.89, 0.86, 0.83, 0.78, 0.67],
            ),
            # Incomplete growing season; Raise a ValueError
            (
                "jones",
                "04-01",
                "11-01",
                "YS-JUL",
                False,
                None,
            ),
        ],
    )
    def test_jones_day_length_latitude_coefficient(self, method, start_date, end_date, freq, floor, results):
        if freq == "YS-JUL":
            setup_dates = {"start_date": "1992-08-01", "end_date": "1993-06-01"}
        else:
            setup_dates = {}

        data = self.data_setup(lats=np.linspace(-65, 65, 13, endpoint=True), **setup_dates)
        if results is None:
            with pytest.raises(ValueError):
                helpers.jones_day_length_latitude_coefficient(
                    dates=data.time,
                    lat=data.lat,
                    start_date=start_date,
                    end_date=end_date,
                    freq=freq,
                    method=method,
                    floor=floor,
                )
        else:
            k = helpers.jones_day_length_latitude_coefficient(
                dates=data.time,
                lat=data.lat,
                start_date=start_date,
                end_date=end_date,
                freq=freq,
                method=method,
                floor=floor,
            )
            np.testing.assert_array_almost_equal(k.transpose()[0], results, 2)

    @pytest.mark.parametrize("constrain", [None, "20 degree_north"])
    def test_gladstones_day_length(self, constrain):
        data = self.data_setup(lats=np.linspace(-65, 65, 13, endpoint=True))
        k = helpers.gladstones_day_length_latitude_coefficient(dates=data.time, lat=data.lat, constrain=constrain)

        events = dict(
            solstice=[
                ["1992-12-21", [1.42, 1.14, 1.03, 0.95, 0.9, 0.85, 1.31, 1.24, 1.17, 1.08, 0.96, 0.77, 0.32]],
                ["1993-06-21", [0.31, 0.77, 0.96, 1.08, 1.17, 1.24, 0.81, 0.85, 0.9, 0.95, 1.03, 1.14, 1.42]],
                ["1993-12-21", [1.42, 1.14, 1.03, 0.95, 0.9, 0.85, 1.31, 1.24, 1.17, 1.08, 0.96, 0.77, 0.32]],
            ],
            equinox=[
                ["1993-03-20", [1.0] * 13]
            ],  # True equinox on 1993-03-20 at 14:41 GMT. Some relative tolerance is needed.
        )

        if constrain == "20 degree_north":
            for entry in events["solstice"]:
                entry[1][5:8] = [1.0] * 3

        for event, evaluations in events.items():
            for e in evaluations:
                if event == "solstice":
                    np.testing.assert_array_almost_equal(k.sel(time=e[0]).transpose(), np.array(e[1]), 2)
                elif event == "equinox":
                    np.testing.assert_allclose(k.sel(time=e[0]).transpose(), np.array(e[1]), rtol=2e-1)


@pytest.mark.parametrize("calendar", ["standard", "noleap"])
def test_cosine_of_solar_zenith_angle(calendar):
    time = xr.date_range("1900-01-01T00:30", "1900-01-03", freq="h", calendar=calendar)
    time = xr.DataArray(time, dims=("time",), coords={"time": time}, name="time")
    lat = xr.DataArray([0, 45, 70], dims=("site",), name="lat", attrs={"units": "degree_north"})
    lon = xr.DataArray([-40, 0, 80], dims=("site",), name="lon", attrs={"units": "degree_east"})
    dec = helpers.solar_declination(time)

    czda = helpers.cosine_of_solar_zenith_angle(time, dec, lat, lon, stat="average", sunlit=True)
    # Data Generated with PyWGBT
    # raw = coszda(
    #     (time + pd.Timedelta('30 m')).data,
    #     convert_units_to(lat, 'rad').data[np.newaxis, :],
    #     convert_units_to(lon, 'rad').data[np.newaxis, :],
    #     1
    # )
    # exp_cza = xr.DataArray(
    #               raw,
    #               dims=('time', 'd', 'site'),
    #               coords={'lat': lat, 'lon': lon, 'time': time}
    #           ).squeeze('d')
    exp_czda = np.array(
        [
            [0.0, 0.0610457, 0.0],
            [0.09999178, 0.18221077, 0.0],
            [0.31387116, 0.285383, 0.0],
            [0.52638271, 0.35026199, 0.0],
            [0.70303168, 0.37242693, 0.0],
        ]
    )
    np.testing.assert_allclose(czda[7:12, :], exp_czda, rtol=1e-3)

    # Same code as above, but with function "cosza".
    cza = helpers.cosine_of_solar_zenith_angle(time, dec, lat, lon, stat="average", sunlit=False)
    exp_cza = np.array(
        [
            [-0.83153798, -0.90358335, -0.34065474],
            [-0.90358299, -0.83874813, -0.26062708],
            [-0.91405234, -0.73561867, -0.18790995],
            [-0.86222963, -0.60121893, -0.12745608],
        ]
    )
    np.testing.assert_allclose(cza[:4, :], exp_cza, rtol=1e-3)


@pytest.mark.parametrize(["in_chunks", "exp_chunks"], [(60, 6 * (2,)), (30, 12 * (1,)), (-1, (12,))])
def test_resample_map(tas_series, in_chunks, exp_chunks):
    pytest.importorskip("flox")
    tas = tas_series(365 * [1]).chunk(time=in_chunks)
    with assert_lazy:
        out = helpers.resample_map(tas, "time", "MS", lambda da: da.mean("time"), map_blocks=True)
    assert out.chunks[0] == exp_chunks
    out.load()  # Trigger compute to see if it actually works


def test_resample_map_dataset(tas_series, pr_series):
    pytest.importorskip("flox")
    tas = tas_series(3 * 365 * [1], start="2000-01-01").chunk(time=365)
    pr = pr_series(3 * 365 * [1], start="2000-01-01").chunk(time=365)
    ds = xr.Dataset({"pr": pr, "tas": tas})
    with set_options(resample_map_blocks=True):
        with assert_lazy:
            out = helpers.resample_map(
                ds,
                "time",
                "YS",
                lambda da: da.mean("time"),
            )
    assert out.chunks["time"] == (1, 1, 1)
    out.load()


def test_resample_map_passthrough(tas_series):
    tas = tas_series(365 * [1])
    with assert_lazy:
        out = helpers.resample_map(tas, "time", "MS", lambda da: da.mean("time"))
    assert not uses_dask(out)


@pytest.mark.parametrize("calendar", [None, "standard"])
def test_make_hourly_temperature(tasmax_series, tasmin_series, calendar):
    tasmax = tasmax_series(np.array([20]), units="degC", calendar=calendar)
    tasmin = tasmin_series(np.array([0]), units="degC", calendar=calendar).expand_dims(lat=[0])

    tasmin.lat.attrs["units"] = "degree_north"
    tas_hourly = helpers.make_hourly_temperature(tasmax, tasmin)
    assert tas_hourly.attrs["units"] == "degC"
    assert tas_hourly.time.size == 24
    expected = np.array(
        [
            0.0,
            3.90180644,
            7.65366865,
            11.11140466,
            14.14213562,
            16.62939225,
            18.47759065,
            19.61570561,
            20.0,
            19.61570561,
            18.47759065,
            16.62939225,
            14.14213562,
            10.32039099,
            8.0848137,
            6.49864636,
            5.26831939,
            4.26306907,
            3.41314202,
            2.67690173,
            2.02749177,
            1.44657476,
            0.92107141,
            0.44132444,
        ]
    )
    np.testing.assert_allclose(tas_hourly.isel(lat=0).values, expected)



"""Simple indice definitions."""

from __future__ import annotations

import xarray

from xclim.core import Quantified
from xclim.core.calendar import select_time
from xclim.core.units import convert_units_to, declare_units, rate2amount, to_agg_units
from xclim.indices.generic import select_resample_op, threshold_count

# Frequencies : YS: year start, QS-DEC: seasons starting in december, MS: month start
# See http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases

# -------------------------------------------------- #
# ATTENTION: ASSUME ALL INDICES WRONG UNTIL TESTED ! #
# -------------------------------------------------- #

__all__ = [
    "frost_days",
    "hot_days",
    "ice_days",
    "max_1day_precipitation_amount",
    "max_n_day_precipitation_amount",
    "max_pr_intensity",
    "sfcWind_max",
    "sfcWind_mean",
    "sfcWind_min",
    "sfcWindmax_max",
    "sfcWindmax_mean",
    "sfcWindmax_min",
    "snow_depth",
    "tg_max",
    "tg_mean",
    "tg_min",
    "tn_max",
    "tn_mean",
    "tn_min",
    "tx_max",
    "tx_mean",
    "tx_min",
]


@declare_units(tas="[temperature]")
def tg_max(tas: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Highest mean temperature.

    The maximum of daily mean temperature.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tas]
        Maximum of daily mean temperature.

    Notes
    -----
    Let :math:`TN_{ij}` be the mean temperature at day :math:`i` of period :math:`j`. Then the maximum
    daily mean temperature for period :math:`j` is:

    .. math::

       TNx_j = max(TN_{ij})
    """
    return tas.resample(time=freq).max(dim="time").assign_attrs(units=tas.units)


@declare_units(tas="[temperature]")
def tg_mean(tas: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Mean of daily average temperature.

    Resample the original daily mean temperature series by taking the mean over each period.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tas]
        The mean daily temperature at the given time frequency.

    Notes
    -----
    Let :math:`TN_i` be the mean daily temperature of day :math:`i`, then for a period :math:`p` starting at
    day :math:`a` and finishing on day :math:`b`:

    .. math::

       TG_p = \frac{\sum_{i=a}^{b} TN_i}{b - a + 1}

    Examples
    --------
    The following would compute for each grid cell of file `tas.day.nc` the mean temperature
    at the seasonal frequency, i.e. DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import tg_mean
    >>> t = xr.open_dataset(path_to_tas_file).tas
    >>> tg = tg_mean(t, freq="QS-DEC")
    """
    return select_resample_op(tas, op="mean", freq=freq)


@declare_units(tas="[temperature]")
def tg_min(tas: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Lowest mean temperature.

    Minimum of daily mean temperature.

    Parameters
    ----------
    tas : xarray.DataArray
        Mean daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tas]
        Minimum of daily minimum temperature.

    Notes
    -----
    Let :math:`TG_{ij}` be the mean temperature at day :math:`i` of period :math:`j`. Then the minimum
    daily mean temperature for period :math:`j` is:

    .. math::

       TGn_j = min(TG_{ij})
    """
    return select_resample_op(tas, op="min", freq=freq)


@declare_units(tasmin="[temperature]")
def tn_max(tasmin: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Highest minimum temperature.

    The maximum of daily minimum temperature.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tasmin]
        Maximum of daily minimum temperature.

    Notes
    -----
    Let :math:`TN_{ij}` be the minimum temperature at day :math:`i` of period :math:`j`. Then the maximum
    daily minimum temperature for period :math:`j` is:

    .. math::

       TNx_j = max(TN_{ij})
    """
    return select_resample_op(tasmin, op="max", freq=freq)


@declare_units(tasmin="[temperature]")
def tn_mean(tasmin: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Mean minimum temperature.

    Mean of daily minimum temperature.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tasmin]
        Mean of daily minimum temperature.

    Notes
    -----
    Let :math:`TN_{ij}` be the minimum temperature at day :math:`i` of period :math:`j`. Then mean
    values in period :math:`j` are given by:

    .. math::

       TN_{ij} = \frac{ \sum_{i=1}^{I} TN_{ij} }{I}
    """
    return select_resample_op(tasmin, op="mean", freq=freq)


@declare_units(tasmin="[temperature]")
def tn_min(tasmin: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Lowest minimum temperature.

    Minimum of daily minimum temperature.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tasmin]
        Minimum of daily minimum temperature.

    Notes
    -----
    Let :math:`TN_{ij}` be the minimum temperature at day :math:`i` of period :math:`j`. Then the minimum
    daily minimum temperature for period :math:`j` is:

    .. math::

       TNn_j = min(TN_{ij})
    """
    return select_resample_op(tasmin, op="min", freq=freq)


@declare_units(tasmax="[temperature]")
def tx_max(tasmax: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Highest max temperature.

    The maximum value of daily maximum temperature.

    Parameters
    ----------
    tasmax : xarray.DataArray
        Maximum daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tasmax]
        Maximum value of daily maximum temperature.

    Notes
    -----
    Let :math:`TX_{ij}` be the maximum temperature at day :math:`i` of period :math:`j`. Then the maximum
    daily maximum temperature for period :math:`j` is:

    .. math::

       TXx_j = max(TX_{ij})
    """
    return select_resample_op(tasmax, op="max", freq=freq)


@declare_units(tasmax="[temperature]")
def tx_mean(tasmax: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Mean max temperature.

    The mean of daily maximum temperature.

    Parameters
    ----------
    tasmax : xarray.DataArray
        Maximum daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tasmax]
        Mean of daily maximum temperature.

    Notes
    -----
    Let :math:`TX_{ij}` be the maximum temperature at day :math:`i` of period :math:`j`. Then mean
    values in period :math:`j` are given by:

    .. math::

       TX_{ij} = \frac{ \sum_{i=1}^{I} TX_{ij} }{I}
    """
    return select_resample_op(tasmax, op="mean", freq=freq)


@declare_units(tasmax="[temperature]")
def tx_min(tasmax: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Lowest max temperature.

    The minimum of daily maximum temperature.

    Parameters
    ----------
    tasmax : xarray.DataArray
        Maximum daily temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as tasmax]
      Minimum of daily maximum temperature.

    Notes
    -----
    Let :math:`TX_{ij}` be the maximum temperature at day :math:`i` of period :math:`j`. Then the minimum
    daily maximum temperature for period :math:`j` is:

    .. math::

       TXn_j = min(TX_{ij})
    """
    return select_resample_op(tasmax, op="min", freq=freq)


@declare_units(tasmax="[temperature]", thresh="[temperature]")
def hot_days(
    tasmax: xarray.DataArray,
    thresh: Quantified = "25 degC",
    freq: str = "YS",
) -> xarray.DataArray:
    r"""
    Hot days index.

    Number of days where daily maximum temperatures are above a threshold temperature.

    Parameters
    ----------
    tasmax : xarray.DataArray
        Maximum daily temperature.
    thresh : Quantified
        Threshold temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [time]
        Hot days index.

    Notes
    -----
    Let :math:`TX_{ij}` be the daily maximum temperature at day :math:`i` of period :math:`j`
    and :math`TT` the threshold. Then counted is the number of days where:

    .. math::

       TX_{ij} > TT
    """
    thresh = convert_units_to(thresh, tasmax)
    out = threshold_count(tasmax, ">", thresh, freq)
    return to_agg_units(out, tasmax, "count", deffreq="D")


@declare_units(tasmin="[temperature]", thresh="[temperature]")
def frost_days(
    tasmin: xarray.DataArray,
    thresh: Quantified = "0 degC",
    freq: str = "YS",
) -> xarray.DataArray:
    r"""
    Frost days index.

    Number of days where daily minimum temperatures are below a threshold temperature.

    Parameters
    ----------
    tasmin : xarray.DataArray
        Minimum daily temperature.
    thresh : Quantified
        Freezing temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [time]
        Frost days index.

    Notes
    -----
    Let :math:`TN_{ij}` be the daily minimum temperature at day :math:`i` of period :math:`j`
    and :math`TT` the threshold. Then counted is the number of days where:

    .. math::

       TN_{ij} < TT
    """
    frz = convert_units_to(thresh, tasmin)
    out = threshold_count(tasmin, "<", frz, freq)
    return to_agg_units(out, tasmin, "count", deffreq="D")


@declare_units(tasmax="[temperature]", thresh="[temperature]")
def ice_days(tasmax: xarray.DataArray, thresh: Quantified = "0 degC", freq: str = "YS") -> xarray.DataArray:
    r"""
    Number of ice/freezing days.

    Number of days when daily maximum temperatures are below a threshold.

    Parameters
    ----------
    tasmax : xarray.DataArray
        Maximum daily temperature.
    thresh : Quantified
        Freezing temperature.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [time]
        Number of ice/freezing days.

    Notes
    -----
    Let :math:`TX_{ij}` be the daily maximum temperature at day :math:`i` of period :math:`j`,
    and :math`TT` the threshold. Then counted is the number of days where:

    .. math::

       TX_{ij} < TT
    """
    frz = convert_units_to(thresh, tasmax)
    out = threshold_count(tasmax, "<", frz, freq)
    return to_agg_units(out, tasmax, "count", deffreq="D")


@declare_units(pr="[precipitation]")
def max_1day_precipitation_amount(pr: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Highest 1-day precipitation amount for a period (frequency).

    Resample the original daily total precipitation temperature series by taking the max over each period.

    Parameters
    ----------
    pr : xarray.DataArray
        Daily precipitation values.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as pr]
        The highest 1-period precipitation flux value at the given time frequency.

    Notes
    -----
    Let :math:`PR_i` be the mean daily precipitation of day `i`, then for a period `j`:

    .. math::

       PRx_{ij} = max(PR_{ij})

    Examples
    --------
    The following would compute for each grid cell the highest 1-day total at an annual frequency:

    >>> from xclim.indices import max_1day_precipitation_amount
    >>> pr = xr.open_dataset(path_to_pr_file).pr
    >>> rx1day = max_1day_precipitation_amount(pr, freq="YS")
    """
    return select_resample_op(pr, op="max", freq=freq)


@declare_units(pr="[precipitation]")
def max_n_day_precipitation_amount(
    pr: xarray.DataArray, window: int = 1, freq: str = "YS", **indexer
) -> xarray.DataArray:
    r"""
    Highest precipitation amount cumulated over a n-day moving window.

    Calculate the n-day rolling sum of the original daily total precipitation series
    and determine the maximum value over each period.

    Parameters
    ----------
    pr : xarray.DataArray
        Daily precipitation values.
    window : int
        Window size in days.
    freq : str
        Resampling frequency.
    **indexer : {dim: indexer}, optional
        Indexing parameters to compute the indicator on a temporal subset of the data.
        The subset is taken after the N-day sum, thus including data from up to ``window -1``
        days before the selected period (and none after).
        It accepts the same arguments as :py:func:`xclim.indices.generic.select_time`.

    Returns
    -------
    xarray.DataArray, [length]
        The highest cumulated n-period precipitation value at the given time frequency.

    Examples
    --------
    The following would compute for each grid cell the highest 5-day total precipitation at an annual frequency:

    >>> from xclim.indices import max_n_day_precipitation_amount
    >>> pr = xr.open_dataset(path_to_pr_file).pr
    >>> out = max_n_day_precipitation_amount(pr, window=5, freq="YS")
    """
    # Rolling sum of the values
    pram = rate2amount(pr)
    arr = pram.rolling(time=window).sum(skipna=False)
    arr = select_time(arr, **indexer)
    return arr.resample(time=freq).max(dim="time").assign_attrs(units=pram.units)


@declare_units(pr="[precipitation]")
def max_pr_intensity(pr: xarray.DataArray, window: int = 1, freq: str = "YS", **indexer) -> xarray.DataArray:
    r"""
    Highest precipitation intensity over a n-hour moving window.

    Calculate the n-hour rolling average of the original hourly total precipitation series
    and determine the maximum value over each period.

    Parameters
    ----------
    pr : xarray.DataArray
        Hourly precipitation values.
    window : int
        Window size in hours.
    freq : str
        Resampling frequency.
    **indexer : {dim: indexer}, optional
        Indexing parameters to compute the indicator on a temporal subset of the data.
        The subset is taken after the N-hour average, thus including data from up to ``window - 1``
        hours before the selected period, and none after.
        It accepts the same arguments as :py:func:`xclim.indices.generic.select_time`.

    Returns
    -------
    xarray.DataArray, [same units as pr]
        The highest cumulated n-hour precipitation intensity at the given time frequency.

    Examples
    --------
    The following would compute the maximum 6-hour precipitation intensity at an annual frequency:

    >>> from xclim.indices import max_pr_intensity
    >>> pr = xr.open_dataset(path_to_pr_file).pr
    >>> out = max_pr_intensity(pr, window=5, freq="YS")
    """
    # Rolling sum of the values
    arr = pr.rolling(time=window).mean(skipna=False)
    arr = select_time(arr, **indexer)
    out = arr.resample(time=freq).max(dim="time")

    out.attrs["units"] = pr.units
    return out


@declare_units(snd="[length]")
def snow_depth(
    snd: xarray.DataArray,
    freq: str = "YS",
) -> xarray.DataArray:
    """
    Mean of daily average snow depth.

    Resample the original daily mean snow depth series by taking the mean over each period.

    Parameters
    ----------
    snd : xarray.DataArray
        Mean daily snow depth.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as snd]
        The mean daily snow depth at the given time frequency.
    """
    return snd.resample(time=freq).mean(dim="time").assign_attrs(units=snd.units)


@declare_units(sfcWind="[speed]")
def sfcWind_max(  # noqa: N802
    sfcWind: xarray.DataArray, freq: str = "YS"
) -> xarray.DataArray:
    r"""
    Highest daily mean wind speed.

    The maximum of daily mean wind speed.

    Parameters
    ----------
    sfcWind : xarray.DataArray
        Mean daily wind speed.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as sfcWind]
        Maximum of daily mean wind speed.

    Notes
    -----
    Let :math:`FG_{ij}` be the mean wind speed at day :math:`i` of period :math:`j`. Then the maximum
    daily mean wind speed for period :math:`j` is:

    .. math::

       FGx_j = max(FG_{ij})

    Examples
    --------
    The following would compute for each grid cell the maximum wind speed
    at the seasonal frequency, i.e. DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import sfcWind_max
    >>> fg = xr.open_dataset(path_to_sfcWind_file).sfcWind
    >>> fg_max = sfcWind_max(fg, freq="QS-DEC")
    """
    return sfcWind.resample(time=freq).max(dim="time").assign_attrs(units=sfcWind.units)


@declare_units(sfcWind="[speed]")
def sfcWind_mean(  # noqa: N802
    sfcWind: xarray.DataArray, freq: str = "YS"
) -> xarray.DataArray:
    r"""
    Mean of daily mean wind speed.

    Resample the original daily mean wind speed series by taking the mean over each period.

    Parameters
    ----------
    sfcWind : xarray.DataArray
        Mean daily wind speed.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as sfcWind]
        The mean daily wind speed at the given time frequency.

    Notes
    -----
    Let :math:`FG_i` be the mean wind speed of day :math:`i`, then for a period :math:`p` starting at
    day :math:`a` and finishing on day :math:`b`:

    .. math::

       FG_m = \frac{\sum_{i=a}^{b} FG_i}{b - a + 1}

    Examples
    --------
    The following would compute for each grid cell the mean wind speed
    at the seasonal frequency, i.e. DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import sfcWind_mean
    >>> fg = xr.open_dataset(path_to_sfcWind_file).sfcWind
    >>> fg_mean = sfcWind_mean(fg, freq="QS-DEC")
    """
    return sfcWind.resample(time=freq).mean(dim="time").assign_attrs(units=sfcWind.units)


@declare_units(sfcWind="[speed]")
def sfcWind_min(  # noqa: N802
    sfcWind: xarray.DataArray, freq: str = "YS"
) -> xarray.DataArray:
    r"""
    Lowest daily mean wind speed.

    The minimum of daily mean wind speed.

    Parameters
    ----------
    sfcWind : xarray.DataArray
        Mean daily wind speed.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as sfcWind]
        Minimum of daily mean wind speed.

    Notes
    -----
    Let :math:`FG_{ij}` be the mean wind speed at day :math:`i` of period :math:`j`. Then the minimum
    daily mean wind speed for period :math:`j` is:

    .. math::

       FGn_j = min(FG_{ij})

    Examples
    --------
    The following would compute for each grid cell the minimum wind speed
    at the seasonal frequency, i.e. DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import sfcWind_min
    >>> fg = xr.open_dataset(path_to_sfcWind_file).sfcWind
    >>> fg_min = sfcWind_min(fg, freq="QS-DEC")
    """
    return sfcWind.resample(time=freq).min(dim="time").assign_attrs(units=sfcWind.units)


@declare_units(sfcWindmax="[speed]")
def sfcWindmax_max(  # noqa: N802
    sfcWindmax: xarray.DataArray, freq: str = "YS"
) -> xarray.DataArray:
    r"""
    Highest maximum wind speed.

    The maximum of daily maximum wind speed.

    Parameters
    ----------
    sfcWindmax : xarray.DataArray
        Maximum daily wind speed.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as sfcWindmax]
        Maximum value of daily maximum wind speed.

    Notes
    -----
    Let :math:`FX_{ij}` be the maximum wind speed at day :math:`i` of period :math:`j`. Then the maximum
    daily maximum wind speed for period :math:`j` is:

    .. math::

       FXx_j = max(FX_{ij})

    Examples
    --------
    The following would compute for each grid cell of the dataset the extreme maximum wind speed
    at the seasonal frequency, i.e. DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import sfcWindmax_max
    >>> max_sfcWindmax = sfcWindmax_max(sfcWindmax_dataset, freq="QS-DEC")
    """
    return sfcWindmax.resample(time=freq).max(dim="time").assign_attrs(units=sfcWindmax.units)


@declare_units(sfcWindmax="[speed]")
def sfcWindmax_mean(  # noqa: N802
    sfcWindmax: xarray.DataArray, freq: str = "YS"
) -> xarray.DataArray:
    r"""
    Mean of daily maximum wind speed.

    Resample the original daily maximum wind speed series by taking the mean over each period.

    Parameters
    ----------
    sfcWindmax : xarray.DataArray
        Maximum daily wind speed.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as sfcWindmax]
        The mean daily maximum wind speed at the given time frequency.

    Notes
    -----
    Let :math:`FX_i` be the maximum wind speed of day :math:`i`, then for a period :math:`p` starting at
    day :math:`a` and finishing on day :math:`b`:

    .. math::

       FX_m = \frac{\sum_{i=a}^{b} FX_i}{b - a + 1}

    Examples
    --------
    The following would compute for each grid cell of the dataset the mean of maximum wind speed
    at the seasonal frequency, i.e. DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import sfcWindmax_mean
    >>> mean_sfcWindmax = sfcWindmax_mean(sfcWindmax_dataset, freq="QS-DEC")
    """
    return sfcWindmax.resample(time=freq).mean(dim="time").assign_attrs(units=sfcWindmax.units)


@declare_units(sfcWindmax="[speed]")
def sfcWindmax_min(  # noqa: N802
    sfcWindmax: xarray.DataArray, freq: str = "YS"
) -> xarray.DataArray:
    r"""
    Lowest daily maximum wind speed.

    The minimum of daily maximum wind speed.

    Parameters
    ----------
    sfcWindmax : xarray.DataArray
        Maximum daily wind speed.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [same units as sfcWindmax]
        Minimum of daily maximum wind speed.

    Notes
    -----
    Let :math:`FX_{ij}` be the maximum wind speed at day :math:`i` of period :math:`j`.
    Then the minimum daily maximum wind speed for period :math:`j` is:

    .. math::

       FXn_j = min(FX_{ij})

    Examples
    --------
    The following would compute for each grid cell of the dataset the minimum of maximum wind speed
    at the seasonal frequency, i.e. DJF, MAM, JJA, SON, DJF, etc.:

    >>> from xclim.indices import sfcWindmax_min
    >>> min_sfcWindmax = sfcWindmax_min(sfcWindmax_dataset, freq="QS-DEC")
    """
    return sfcWindmax.resample(time=freq).min(dim="time").assign_attrs(units=sfcWindmax.units)



"""
Atmospheric Indicators
======================

While the `indices` module stores the computing functions, this module defines Indicator classes and instances that
include a number of functionalities, such as input validation, unit conversion, output meta-data handling,
and missing value masking.

The concept followed here is to define Indicator subclasses for each input variable, then create instances
for each indicator.
"""

from __future__ import annotations

import functools
import warnings

from xclim.indicators import convert
from xclim.indicators.convert._conversion import __all__ as _conversion_all  # noqa: F401

from ._precip import *
from ._precip import __all__ as _precip_all
from ._synoptic import *
from ._synoptic import __all__ as _synoptic_all
from ._temperature import *
from ._temperature import __all__ as _temperature_all
from ._wind import *
from ._wind import __all__ as _wind_all

__all__ = _precip_all + _synoptic_all + _temperature_all + _wind_all


def _deprecated_alias(func_name):
    """Factory to create a deprecated alias for a function in new_module."""
    new_func = getattr(convert, func_name)

    @functools.wraps(new_func)
    def wrapper(*args, **kwargs):
        warnings.warn(
            f"{func_name} is deprecated and will be removed in a future release. "
            f"Use xclim.convert.{func_name} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return new_func(*args, **kwargs)

    return wrapper


for _name in _conversion_all:
    # Exclude snd_to_snw and snw_to_snd as they were never in atmos module
    if _name not in ["snd_to_snw", "snw_to_snd"]:
        continue
    _obj = getattr(convert, _name)
    if callable(_obj):
        globals()[_name] = _deprecated_alias(_name)



# This File is for testing atmos indicators that fit neither in test_precip or test_temperature
# Expected values might be the same as for the indices tests, see test_indices comments.
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr

from xclim import convert, set_options
from xclim.core.units import convert_units_to
from xclim.indices.converters import shortwave_downwelling_radiation_from_clearness_index
from xclim.indices.helpers import extraterrestrial_solar_radiation

K2C = 273.16


def test_wind_speed_from_vectors():
    uas = xr.DataArray(np.array([3.0, -3.0]), dims=["x"])
    uas.attrs["units"] = "m s-1"
    vas = xr.DataArray(np.array([4.0, -4.0]), dims=["x"])
    vas.attrs["units"] = "m s-1"

    wind, wind_dir = convert.wind_speed_from_vector(uas=uas, vas=vas)
    np.testing.assert_allclose(wind, [5.0, 5.0])
    np.testing.assert_allclose(wind_dir, [216.86989764584402, 36.86989764584402])

    # missing values
    uas[0] = np.nan
    wind, wind_dir = convert.wind_speed_from_vector(uas=uas, vas=vas)
    np.testing.assert_array_equal(wind.isnull(), [True, False])
    np.testing.assert_array_equal(wind_dir.isnull(), [True, False])

    # Calm thresh and northerly
    uas[:] = 0
    vas[0] = 0.9
    vas[1] = -1.1
    wind, wind_dir = convert.wind_speed_from_vector(uas=uas, vas=vas, calm_wind_thresh="1 m/s")
    np.testing.assert_array_equal(wind, [0.9, 1.1])
    np.testing.assert_allclose(wind_dir, [0.0, 360.0])


def test_wind_vector_from_speed():
    sfcWind = xr.DataArray(np.array([3.0, 5.0, 0.2]), dims=["x"])
    sfcWind.attrs["units"] = "m s-1"
    sfcWindfromdir = xr.DataArray(np.array([360.0, 36.86989764584402, 0.0]), dims=["x"])
    sfcWindfromdir.attrs["units"] = "degree"

    uas, vas = convert.wind_vector_from_speed(sfcWind=sfcWind, sfcWindfromdir=sfcWindfromdir)
    np.testing.assert_allclose(uas, [0.0, -3.0, 0.0], atol=1e-14)
    np.testing.assert_allclose(vas, [-3.0, -4.0, -0.2], atol=1e-14)

    # missing values
    sfcWind[0] = np.nan
    sfcWindfromdir[1] = np.nan
    wind_vectors = convert.wind_vector_from_speed(sfcWind=sfcWind, sfcWindfromdir=sfcWindfromdir)
    # try using namedtuple accessor
    np.testing.assert_array_equal(wind_vectors.uas.isnull(), [True, True, False])
    np.testing.assert_array_equal(wind_vectors.vas.isnull(), [True, True, False])


def test_relative_humidity_dewpoint(timeseries):
    np.testing.assert_allclose(
        convert.relative_humidity_from_dewpoint(
            tas=timeseries(np.array([-20, -10, -1, 10, 20, 25, 30, 40, 60]) + K2C, "tas"),
            tdps=timeseries(np.array([-15, -10, -2, 5, 10, 20, 29, 20, 30]) + K2C, "tdps"),
        ),
        timeseries([np.nan, 100, 93, 71, 52, 73, 94, 31, 20], "hurs"),
        rtol=0.02,
        atol=1,
    )


def test_relative_humidity_dewpoint_clip(timeseries):
    np.testing.assert_allclose(
        convert.relative_humidity_from_dewpoint(
            tas=timeseries(np.array([-20, -10, -1, 10, 20, 25, 30, 40, 60]) + K2C, "tas"),
            tdps=timeseries(np.array([-15, -10, -2, 5, 10, 20, 29, 20, 30]) + K2C, "tdps"),
            invalid_values="clip",
        ),
        timeseries([100, 100, 93, 71, 52, 73, 94, 31, 20], "hurs"),
        rtol=0.02,
        atol=1,
    )


def test_humidex(tas_series):
    tas = tas_series([15, 25, 35, 40])
    tas.attrs["units"] = "C"

    dtas = tas_series([10, 15, 25, 25])
    dtas.attrs["units"] = "C"

    # expected values from https://en.wikipedia.org/wiki/Humidex
    h = convert.humidex(tas, dtas)
    np.testing.assert_array_almost_equal(h, [16, 29, 47, 52], 0)
    assert h.name == "humidex"

    # Test with dataset (#1432)
    ds = xr.Dataset({"tas": tas, "dewpoint": dtas})
    h2 = convert.humidex(ds=ds, tdps="dewpoint")
    np.testing.assert_array_almost_equal(h, h2)


def test_heat_index(atmosds):
    # Keep just Montreal values for summer as we need tas > 20 degC
    tas = atmosds.tasmax[1][150:170]
    hurs = atmosds.hurs[1][150:170]

    expected = np.array(
        [
            25.0,
            27.0,
            29.0,
            27.0,
            24.0,
            np.nan,
            np.nan,
            23.0,
            24.0,
            np.nan,
            np.nan,
            24.0,
            28.0,
            25.0,
            30.0,
            26.0,
            31.0,
            33.0,
            34.0,
            28.0,
        ]
    )

    with set_options(cf_compliance="raise"):
        hi = convert.heat_index(tas, hurs)
    np.testing.assert_array_almost_equal(hi, expected, 0)
    assert hi.name == "heat_index"


def test_saturation_vapor_pressure(tas_series):
    tas = tas_series(np.array([-20, -10, -1, 10, 20, 25, 30, 40, 60]) + K2C)
    e_sat_exp = [103, 260, 563, 1228, 2339, 3169, 4247, 7385, 19947]
    e_sat = convert.saturation_vapor_pressure(
        tas=tas,
        method="sonntag90",
        ice_thresh="0 degC",
    )
    np.testing.assert_allclose(e_sat, e_sat_exp, atol=0.5, rtol=0.005)
    assert e_sat.name == "e_sat"


def test_relative_humidity(tas_series, hurs_series, huss_series, ps_series):
    tas = tas_series(np.array([-10, -10, 10, 20, 35, 50, 75, 95]) + K2C)
    hurs_exp = hurs_series([np.nan, 63.0, 66.0, 34.0, 14.0, 6.0, 1.0, 0.0])
    ps = ps_series([101325] * 8)
    huss = huss_series([0.003, 0.001] + [0.005] * 7)

    hurs = convert.relative_humidity(
        tas=tas,
        huss=huss,
        ps=ps,
        method="sonntag90",
        ice_thresh="0 degC",
    )
    np.testing.assert_allclose(hurs, hurs_exp, atol=0.5, rtol=0.005)
    assert hurs.name == "hurs"


def test_specific_humidity(tas_series, hurs_series, huss_series, ps_series):
    tas = tas_series(np.array([20, -10, 10, 20, 35, 50, 75, 95]) + K2C)
    hurs = hurs_series([150, 10, 90, 20, 80, 50, 70, 40, 30])
    ps = ps_series(1000 * np.array([100] * 4 + [101] * 4))
    huss_exp = huss_series([np.nan, 1.6e-4, 6.9e-3, 3.0e-3, 2.9e-2, 4.1e-2, 2.1e-1, 5.7e-1])

    huss = convert.specific_humidity(
        tas=tas,
        hurs=hurs,
        ps=ps,
        method="sonntag90",
        ice_thresh="0 degC",
    )
    np.testing.assert_allclose(huss, huss_exp, atol=1e-4, rtol=0.05)
    assert huss.name == "huss"


def test_specific_humidity_from_dewpoint(tas_series, ps_series, huss_series):
    tdps = tas_series([272, 283, 293])
    ps = ps_series([100000, 105000, 110000])
    # Computed from MetPy
    # >>> from metpy.units import units as u
    # >>> from metpy.calc import specific_humidity_from_dewpoint as sh
    # >>> sh([100000, 105000, 110000] * u.Pa, [272, 283, 293] * u.degK)
    # array([0.0035031, 0.00722795, 0.01319614]) < Unit('dimensionless') >
    huss_exp = huss_series([0.0035031, 0.00722795, 0.01319614])

    huss = convert.specific_humidity_from_dewpoint(
        tdps=tdps,
        ps=ps,
        method="sonntag90",
    )
    np.testing.assert_allclose(huss, huss_exp, atol=1e-4, rtol=0.05)
    assert huss.name == "huss"


@pytest.mark.parametrize("method", ["tetens30", "wmo08", "aerk96", "buck81"])
def test_dewpoint_from_specific_humidity(atmosds, method):
    tdps = convert.dewpoint_from_specific_humidity(huss=atmosds.huss, ps=atmosds.ps)
    # atmosds.huss was computed with Sonntag90, so we can't expect a perfect match
    xr.testing.assert_allclose(tdps, atmosds.tdps, atol=0.2, rtol=0.001)


def test_snowfall_approximation(pr_series, tas_series, prsn_series):
    # Data from CanDCS-M6 : lon=-73, lat=45.5, time=('2001-01-01', '2001-01-20')
    pr = pr_series(
        [
            4.83717,
            1.89218,
            3.21972,
            10.75767,
            2.25840,
            10.29990,
            13.13808,
            17.63949,
            16.11359,
            0.00006,
            27.93933,
            8.95711,
            9.33858,
            2.25840,
            1.19027,
            0.00006,
            0.00006,
            0.00006,
            9.32332,
            22.98015,
        ],
    )
    tas = tas_series(
        [
            -5.08431,
            -6.09445,
            -6.73533,
            -3.04570,
            -7.70428,
            -7.64477,
            3.10979,
            3.02281,
            -5.82589,
            -11.81048,
            -7.29076,
            0.62867,
            -7.70428,
            -18.71214,
            -7.58678,
            -1.72427,
            -0.09003,
            0.24414,
            -0.40894,
            3.72473,
        ],
        units="°C",
    )
    exp = prsn_series(
        [
            4.83717,
            1.89218,
            3.21972,
            10.72715,
            2.25840,
            10.29990,
            0.74775,
            1.12923,
            16.11359,
            0.00006,
            27.93933,
            6.13419,
            9.33858,
            2.25840,
            1.19027,
            0.00006,
            0.00006,
            0.00006,
            8.45356,
            0.56465,
        ]
    )
    prsn = convert.snowfall_approximation(pr, tas=tas, method="dai_annual", clip_temp="20 °C")
    np.testing.assert_allclose(prsn, exp, atol=1e-4, rtol=1e-2)


def test_rain_approximation(pr_series, tas_series):
    pr = pr_series(np.ones(10))
    tas = tas_series(np.arange(10) + K2C)

    prlp = convert.rain_approximation(pr, tas=tas, thresh="5 degC", method="binary")

    np.testing.assert_allclose(prlp, [0, 0, 0, 0, 0, 1, 1, 1, 1, 1], atol=1e-5, rtol=1e-3)


def test_wind_chill_index(atmosds):
    out = convert.wind_chill_index(ds=atmosds)

    np.testing.assert_allclose(out.isel(time=0), [np.nan, -6.716, -35.617, -8.486, np.nan], rtol=1e-3)

    out_us = convert.wind_chill_index(ds=atmosds, method="US")

    np.testing.assert_allclose(out_us.isel(time=0), [-1.429, -6.716, -35.617, -8.486, 2.781], rtol=1e-3)


def test_wind_profile(atmosds):
    out: xr.DataArray = convert.wind_profile(wind_speed=atmosds.sfcWind, h_r="10 m", h="100 m", alpha=1 / 7)
    assert out.attrs["units"] == "m s-1"
    assert (out > atmosds.sfcWind).all()


def test_wind_power_potential(atmosds):
    out = convert.wind_power_potential(wind_speed=atmosds.sfcWind)
    assert out.attrs["units"] == "1"
    assert (out >= 0).all()
    assert (out <= 1).all()


def test_wind_power_potential_from_3h_series():
    """Test a typical computation workflow from 3-hourly time series to daily power production in MWh."""
    from xclim.core.units import convert_units_to
    from xclim.indices.generic import select_resample_op
    from xclim.testing.helpers import test_timeseries

    w = test_timeseries(np.ones(96) * 15, variable="sfcWind", start="7/1/2000", units="m s-1", freq="3h")
    out = convert.wind_power_potential(wind_speed=w)

    # Multiply with nominal capacity
    power = out * 100
    power.attrs["units"] = "MW"
    annual_power: xr.DataArray = convert_units_to(select_resample_op(power, op="integral", freq="D"), "MWh")
    assert (annual_power == 100 * 24).all()


class TestPotentialEvapotranspiration:
    def test_convert_units(self, atmosds):
        ds = atmosds

        tn = ds.tasmin
        tx = ds.tasmax
        tm = ds.tas
        hurs = ds.hurs
        rsds = ds.rsds
        rsus = ds.rsus
        rlds = ds.rlds
        rlus = ds.rlus
        uas = ds.uas
        vas = ds.vas

        sfcWind, _ = convert.wind_speed_from_vector(uas, vas)

        with xr.set_options(keep_attrs=True):
            tnC = tn - K2C
            tnC.attrs["units"] = "degC"
            tmC = tm - K2C
            tmC.attrs["units"] = "degC"
            hurs_pct = hurs * 100
            hurs_pct.attrs["units"] = "%"

        pet_br65 = convert.potential_evapotranspiration(tn, tx, method="BR65")
        pet_br65C = convert.potential_evapotranspiration(tnC, tx, method="BR65")
        pet_hg85 = convert.potential_evapotranspiration(tn, tx, method="HG85")
        pet_hg85C = convert.potential_evapotranspiration(tnC, tx, method="HG85")
        pet_tw48 = convert.potential_evapotranspiration(tas=tm, method="TW48")
        pet_tw48C = convert.potential_evapotranspiration(tas=tmC, method="TW48")
        pet_mb05 = convert.potential_evapotranspiration(tn, tx, method="MB05")
        pet_mb05C = convert.potential_evapotranspiration(tnC, tx, method="MB05")
        pet_fao_pm98 = convert.potential_evapotranspiration(
            tn,
            tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )
        pet_fao_pm98C = convert.potential_evapotranspiration(
            tnC,
            tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )
        pet_fao_pm98pct = convert.potential_evapotranspiration(
            tn,
            tx,
            hurs=hurs_pct,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )

        np.testing.assert_allclose(pet_br65, pet_br65C, atol=1)
        np.testing.assert_allclose(pet_hg85, pet_hg85C, atol=1)
        np.testing.assert_allclose(pet_tw48, pet_tw48C, atol=1)
        np.testing.assert_allclose(pet_mb05, pet_mb05C, atol=1)
        np.testing.assert_allclose(pet_fao_pm98, pet_fao_pm98C, atol=1)
        np.testing.assert_allclose(pet_fao_pm98, pet_fao_pm98pct, atol=1)

    def test_nan_values(self, atmosds):
        ds = atmosds

        tn = ds.tasmin
        tx = ds.tasmax
        tm = ds.tas
        hurs = ds.hurs
        rsds = ds.rsds
        rsus = ds.rsus
        rlds = ds.rlds
        rlus = ds.rlus
        uas = ds.uas
        vas = ds.vas

        sfcWind, _ = convert.wind_speed_from_vector(uas, vas)

        tn[0, 100] = np.nan
        tx[0, 101] = np.nan

        pet_br65 = convert.potential_evapotranspiration(tn, tx, method="BR65")
        pet_hg85 = convert.potential_evapotranspiration(tn, tx, method="HG85")
        pet_fao_pm98 = convert.potential_evapotranspiration(
            tn,
            tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )

        tm[0, 0:31] = np.nan

        pet_tw48 = convert.potential_evapotranspiration(tas=tm, method="TW48")

        np.testing.assert_allclose(pet_br65.isel(location=0, time=slice(100, 102)), [np.nan, np.nan])
        np.testing.assert_allclose(pet_hg85.isel(location=0, time=slice(100, 102)), [np.nan, np.nan])
        np.testing.assert_allclose(
            pet_fao_pm98.isel(location=0, time=slice(100, 102)),
            [np.nan, np.nan],
        )
        np.testing.assert_allclose(pet_tw48.isel(location=0, time=0), [np.nan])


class TestWaterBudget:
    def test_convert_units(self, atmosds):
        ds = atmosds

        tn = ds.tasmin
        tx = ds.tasmax
        pr = ds.pr
        pet = ds.evspsblpot
        hurs = ds.hurs
        rsds = ds.rsds
        rsus = ds.rsus
        rlds = ds.rlds
        rlus = ds.rlus
        uas = ds.uas
        vas = ds.vas

        sfcWind, _ = convert.wind_speed_from_vector(uas, vas)

        with xr.set_options(keep_attrs=True):
            tnC = tn - K2C
            tnC.attrs["units"] = "degC"
            prR = pr * 86400
            prR.attrs["units"] = "mm/day"
            petR = pet * 86400
            petR.attrs["units"] = "mm/day"

        p_pet_br65 = convert.water_budget_from_tas(pr, tasmin=tn, tasmax=tx, method="BR65")
        p_pet_br65C = convert.water_budget_from_tas(prR, tasmin=tnC, tasmax=tx, method="BR65")
        p_pet_hg85 = convert.water_budget_from_tas(pr, tasmin=tn, tasmax=tx, method="HG85")
        p_pet_hg85C = convert.water_budget_from_tas(prR, tasmin=tnC, tasmax=tx, method="HG85")
        p_pet_tw48 = convert.water_budget_from_tas(pr, tasmin=tn, tasmax=tx, method="TW48")
        p_pet_tw48C = convert.water_budget_from_tas(prR, tasmin=tnC, tasmax=tx, method="TW48")

        p_pet_fao_pm98 = convert.water_budget_from_tas(
            pr=pr,
            tasmin=tn,
            tasmax=tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )
        p_pet_fao_pm98R = convert.water_budget_from_tas(
            pr=prR,
            tasmin=tn,
            tasmax=tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )

        p_pet_evpot = convert.water_budget(pr, evspsblpot=pet)
        p_pet_evpotR = convert.water_budget(prR, evspsblpot=petR)

        np.testing.assert_allclose(p_pet_br65, p_pet_br65C, atol=1)
        np.testing.assert_allclose(p_pet_hg85, p_pet_hg85C, atol=1)
        np.testing.assert_allclose(p_pet_tw48, p_pet_tw48C, atol=1)
        np.testing.assert_allclose(p_pet_fao_pm98, p_pet_fao_pm98R, atol=1)
        np.testing.assert_allclose(p_pet_evpot, p_pet_evpotR, atol=1)

    def test_nan_values(self, atmosds):
        ds = atmosds

        tn = ds.tasmin
        tx = ds.tasmax
        tm = ds.tas
        pr = ds.pr
        pet = ds.evspsblpot
        hurs = ds.hurs
        rsds = ds.rsds
        rsus = ds.rsus
        rlds = ds.rlds
        rlus = ds.rlus
        uas = ds.uas
        vas = ds.vas

        sfcWind, _ = convert.wind_speed_from_vector(uas, vas)

        tn[0, 100] = np.nan
        tx[0, 101] = np.nan

        p_pet_br65 = convert.water_budget_from_tas(pr, tasmin=tn, tasmax=tx, method="BR65")
        p_pet_hg85 = convert.water_budget_from_tas(pr, tasmin=tn, tasmax=tx, method="HG85")
        p_pet_fao_pm98 = convert.water_budget_from_tas(
            pr=pr,
            tasmin=tn,
            tasmax=tx,
            hurs=hurs,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            sfcWind=sfcWind,
            method="FAO_PM98",
        )

        tm[0, 0:31] = np.nan

        p_pet_tw48 = convert.water_budget_from_tas(pr, tas=tm, method="TW48")

        pet[0, 0:31] = np.nan

        p_pet_evpot = convert.water_budget(pr, evspsblpot=pet)

        np.testing.assert_allclose(p_pet_br65[0, 100:102], [np.nan, np.nan])
        np.testing.assert_allclose(p_pet_hg85[0, 100:102], [np.nan, np.nan])
        np.testing.assert_allclose(p_pet_fao_pm98[0, 100:102], [np.nan, np.nan])
        np.testing.assert_allclose(p_pet_tw48[0, 0], [np.nan])
        np.testing.assert_allclose(p_pet_evpot[0, 0], [np.nan])


class TestUTCI:
    def test_universal_thermal_climate_index(self, atmosds):
        dataset = atmosds

        tas = dataset.tas
        hurs = dataset.hurs
        sfcWind, _ = convert.wind_speed_from_vector(uas=dataset.uas, vas=dataset.vas)
        rsds = dataset.rsds
        rsus = dataset.rsus
        rlds = dataset.rlds
        rlus = dataset.rlus
        # Expected values
        utci_exp = [253.887, 254.367, 238.196, 252.826, 264.495]

        utci = convert.universal_thermal_climate_index(
            tas=tas,
            hurs=hurs,
            sfcWind=sfcWind,
            rsds=rsds,
            rsus=rsus,
            rlds=rlds,
            rlus=rlus,
            stat="sunlit",
        )

        np.testing.assert_allclose(utci.isel(time=0), utci_exp, rtol=1e-03)


class TestMeanRadiantTemperature:
    def test_mean_radiant_temperature(self, atmosds):
        dataset = atmosds

        rsds = dataset.rsds
        rsus = dataset.rsus
        rlds = dataset.rlds
        rlus = dataset.rlus

        # Expected values
        exp_sun = [276.911, 274.742, 243.202, 268.012, 278.505]
        exp_ins = [276.911, 274.742, 243.202, 268.012, 278.505]

        mrt_sun = convert.mean_radiant_temperature(rsds, rsus, rlds, rlus, stat="sunlit")
        mrt_ins = convert.mean_radiant_temperature(rsds, rsus, rlds, rlus, stat="instant")

        rtol = 1e-03
        np.testing.assert_allclose(mrt_sun.isel(time=0), exp_sun, rtol=rtol)
        np.testing.assert_allclose(mrt_ins.isel(time=0), exp_ins, rtol=rtol)


class TestClearnessIndex:
    def test_ci_physical_constraints(self, atmosds):
        rsds = atmosds.rsds
        rtop = extraterrestrial_solar_radiation(rsds.time, rsds.lat)
        ci = convert.clearness_index(rsds)
        assert ((ci >= 0) | (ci <= rtop) | (ci.isnull() & rsds.isnull())).all()

    def test_ci_and_inverse_transform(self, atmosds):
        rsds = atmosds.rsds
        ci = convert.clearness_index(rsds)
        rsds2 = shortwave_downwelling_radiation_from_clearness_index(ci)
        rsds2 = convert_units_to(rsds2, rsds)
        assert not ((rsds.isnull() ^ rsds2.isnull()).all())
        np.testing.assert_allclose(rsds.values, rsds2.transpose(*rsds.dims).values)



"""Hydrological indice definitions."""

from __future__ import annotations

import numpy as np
import xarray
from scipy.stats import rv_continuous

from xclim.core._types import DateStr, Quantified
from xclim.core.calendar import get_calendar
from xclim.core.missing import at_least_n_valid
from xclim.core.units import convert_units_to, declare_units, rate2amount, to_agg_units
from xclim.indices.generic import threshold_count
from xclim.indices.stats import standardized_index

from . import generic

__all__ = [
    "antecedent_precipitation_index",
    "base_flow_index",
    "flow_index",
    "high_flow_frequency",
    "low_flow_frequency",
    "melt_and_precip_max",
    "rb_flashiness_index",
    "snd_max",
    "snd_max_doy",
    "snow_melt_we_max",
    "snw_max",
    "snw_max_doy",
    "standardized_groundwater_index",
    "standardized_streamflow_index",
]


@declare_units(q="[discharge]")
def base_flow_index(q: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Base flow index.

    Return the base flow index, defined as the minimum 7-day average flow divided by the mean flow.

    Parameters
    ----------
    q : xarray.DataArray
        Rate of river discharge.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [dimensionless]
        Base flow index.

    Notes
    -----
    Let :math:`\mathbf{q}=q_0, q_1, \ldots, q_n` be the sequence of daily discharge and :math:`\overline{\mathbf{q}}`
    the mean flow over the period. The base flow index is given by:

    .. math::

       \frac{\min(\mathrm{CMA}_7(\mathbf{q}))}{\overline{\mathbf{q}}}

    where :math:`\mathrm{CMA}_7` is the seven days moving average of the daily flow:

    .. math::

       \mathrm{CMA}_7(q_i) = \frac{\sum_{j=i-3}^{i+3} q_j}{7}
    """
    m7 = q.rolling(time=7, center=True).mean(skipna=False).resample(time=freq)
    mq = q.resample(time=freq)

    m7m = m7.min(dim="time")
    out = m7m / mq.mean(dim="time")
    out.attrs["units"] = ""
    return out


@declare_units(q="[discharge]")
def rb_flashiness_index(q: xarray.DataArray, freq: str = "YS") -> xarray.DataArray:
    r"""
    Richards-Baker flashiness index.

    Measures oscillations in flow relative to total flow, quantifying the frequency and rapidity of short term changes
    in flow, based on :cite:t:`baker_new_2004`.

    Parameters
    ----------
    q : xarray.DataArray
        Rate of river discharge.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray, [dimensionless]
        R-B Index.

    Notes
    -----
    Let :math:`\mathbf{q}=q_0, q_1, \ldots, q_n` be the sequence of daily discharge, the R-B Index is given by:

    .. math::

       \frac{\sum_{i=1}^n |q_i - q_{i-1}|}{\sum_{i=1}^n q_i}

    References
    ----------
    :cite:cts:`baker_new_2004`
    """
    d = np.abs(q.diff(dim="time")).resample(time=freq)
    mq = q.resample(time=freq)
    out = d.sum(dim="time") / mq.sum(dim="time")
    out.attrs["units"] = ""
    return out


@declare_units(
    q="[discharge]",
    params="[]",
)
def standardized_streamflow_index(
    q: xarray.DataArray,
    freq: str | None = "MS",
    window: int = 1,
    dist: str | rv_continuous = "genextreme",
    method: str = "ML",
    fitkwargs: dict | None = None,
    cal_start: DateStr | None = None,
    cal_end: DateStr | None = None,
    params: Quantified | None = None,
    **indexer,
) -> xarray.DataArray:
    r"""
    Standardized Streamflow Index (SSI).

    Parameters
    ----------
    q : xarray.DataArray
        Rate of river discharge.
    freq : str, optional
        Resampling frequency. A monthly or daily frequency is expected. Option `None` assumes
        that the desired resampling has already been applied input dataset and will skip the resampling step.
    window : int
        Averaging window length relative to the resampling frequency. For example, if `freq="MS"`,
        i.e. a monthly resampling, the window is an integer number of months.
    dist : {"genextreme", "fisk"} or `rv_continuous` function
        Name of the univariate distribution, or a callable `rv_continuous` (see :py:mod:`scipy.stats`).
    method : {"APP", "ML", "PWM"}
        Name of the fitting method, such as `ML` (maximum likelihood), `APP` (approximate). The approximate method
        uses a deterministic function that does not involve any optimization, which can be sensitive to noise.
        `PWM` should be used with a `lmoments3` distribution.
    fitkwargs : dict, optional
        Kwargs passed to ``xclim.indices.stats.fit`` used to impose values of certain parameters (`floc`, `fscale`).
    cal_start : DateStr, optional
        Start date of the calibration period. A `DateStr` is expected, that is a `str` in format `"YYYY-MM-DD"`.
        Default option `None` means that the calibration period begins at the start of the input dataset.
    cal_end : DateStr, optional
        End date of the calibration period. A `DateStr` is expected, that is a `str` in format `"YYYY-MM-DD"`.
        Default option `None` means that the calibration period finishes at the end of the input dataset.
    params : xarray.DataArray, optional
        Fit parameters.
        The `params` can be computed using ``xclim.indices.stats.standardized_index_fit_params`` in advance.
        The output can be given here as input, and it overrides other options.
    **indexer : Indexer
        Indexing parameters to compute the indicator on a temporal subset of the data.
        It accepts the same arguments as :py:func:`xclim.indices.generic.select_time`.

    Returns
    -------
    xarray.DataArray, [unitless]
        Standardized Streamflow Index.

    See Also
    --------
    xclim.indices._agro.standardized_precipitation_index : Standardized Precipitation Index.
    xclim.indices.stats.standardized_index : Standardized Index.
    xclim.indices.stats.standardized_index_fit_params : Standardized Index Fit Params.

    Notes
    -----
    * N-month SSI / N-day SSI is determined by choosing the `window = N` and the appropriate frequency `freq`.
    * Supported statistical distributions are: ["genextreme", "fisk"], where "fisk" is scipy's implementation of
       a log-logistic distribution.
    * If `params` is provided, it overrides the `cal_start`, `cal_end`, `freq`, `window`, `dist`, and `method` options.
    * "APP" method only supports two-parameter distributions. Parameter `loc` needs to be fixed to use method "APP".
    * The standardized index is bounded by ±8.21. 8.21 is the largest standardized index as constrained by the
      float64 precision in the inversion to the normal distribution.

    References
    ----------
    :cite:cts:`vicente-serrano_2012`

    Examples
    --------
    >>> from datetime import datetime
    >>> from xclim.indices import standardized_streamflow_index
    >>> ds = xr.open_dataset(path_to_q_file)
    >>> q = ds.q_sim
    >>> cal_start, cal_end = "2006-05-01", "2008-06-01"
    >>> ssi_3 = standardized_streamflow_index(
    ...     q,
    ...     freq="MS",
    ...     window=3,
    ...     dist="genextreme",
    ...     method="ML",
    ...     cal_start=cal_start,
    ...     cal_end=cal_end,
    ... )  # Computing SSI-3 months using a GEV distribution for the fit
    >>> # Fitting parameters can also be obtained first, then reused as input.
    >>> from xclim.indices.stats import standardized_index_fit_params
    >>> params = standardized_index_fit_params(
    ...     q.sel(time=slice(cal_start, cal_end)),
    ...     freq="MS",
    ...     window=3,
    ...     dist="genextreme",
    ...     method="ML",
    ... )  # First getting params
    >>> ssi_3 = standardized_streamflow_index(q, params=params)
    """
    fitkwargs = fitkwargs or {}
    dist_methods = {"genextreme": ["ML", "APP"], "fisk": ["ML", "APP"]}
    if isinstance(dist, str):
        if dist in dist_methods:
            if method not in dist_methods[dist]:
                raise NotImplementedError(f"{method} method is not implemented for {dist} distribution")
        else:
            raise NotImplementedError(f"{dist} distribution is not yet implemented.")

    zero_inflated = False
    ssi = standardized_index(
        q,
        freq=freq,
        window=window,
        dist=dist,
        method=method,
        zero_inflated=zero_inflated,
        fitkwargs=fitkwargs,
        cal_start=cal_start,
        cal_end=cal_end,
        params=params,
        **indexer,
    )

    return ssi


@declare_units(snd="[length]")
def snd_max(snd: xarray.DataArray, freq: str = "YS-JUL") -> xarray.DataArray:
    """
    Maximum snow depth.

    The maximum daily snow depth.

    Parameters
    ----------
    snd : xarray.DataArray
        Snow depth (mass per area).
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The maximum snow depth over a given number of days for each period. [length].
    """
    return generic.select_resample_op(snd, op="max", freq=freq)


@declare_units(snd="[length]")
def snd_max_doy(snd: xarray.DataArray, freq: str = "YS-JUL") -> xarray.DataArray:
    """
    Day of year of maximum snow depth.

    Day of year when surface snow reaches its peak value. If snow depth is 0 over entire period, return NaN.

    Parameters
    ----------
    snd : xarray.DataArray
        Surface snow depth.
    freq : str
         Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The day of year at which snow depth reaches its maximum value.
    """
    # Identify periods where there is at least one non-null value for snow depth
    valid = at_least_n_valid(snd.where(snd > 0), n=1, freq=freq)

    # Compute doymax. Will return first time step if all snow depths are 0.
    out = generic.select_resample_op(snd.where(snd > 0, 0), op=generic.doymax, freq=freq)
    out.attrs.update(units="", is_dayofyear=np.int32(1), calendar=get_calendar(snd))

    # Mask arrays that miss at least one non-null snd.
    return out.where(~valid)


@declare_units(snw="[mass]/[area]")
def snw_max(snw: xarray.DataArray, freq: str = "YS-JUL") -> xarray.DataArray:
    """
    Maximum snow amount.

    The maximum daily snow amount.

    Parameters
    ----------
    snw : xarray.DataArray
        Snow amount (mass per area).
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The maximum snow amount over a given number of days for each period. [mass/area].
    """
    return generic.select_resample_op(snw, op="max", freq=freq)


@declare_units(snw="[mass]/[area]")
def snw_max_doy(snw: xarray.DataArray, freq: str = "YS-JUL") -> xarray.DataArray:
    """
    Day of year of maximum snow amount.

    Day of year when surface snow amount reaches its peak value. If snow amount is 0 over entire period, return NaN.

    Parameters
    ----------
    snw : xarray.DataArray
        Surface snow amount.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The day of year at which snow amount reaches its maximum value.
    """
    # Identify periods where there is at least one non-null value for snow depth
    valid = at_least_n_valid(snw.where(snw > 0), n=1, freq=freq)

    # Compute doymax. Will return first time step if all snow depths are 0.
    out = generic.select_resample_op(snw.where(snw > 0, 0), op=generic.doymax, freq=freq)
    out.attrs.update(units="", is_dayofyear=np.int32(1), calendar=get_calendar(snw))

    # Mask arrays that miss at least one non-null snd.
    return out.where(~valid)


@declare_units(snw="[mass]/[area]")
def snow_melt_we_max(snw: xarray.DataArray, window: int = 3, freq: str = "YS-JUL") -> xarray.DataArray:
    """
    Maximum snow melt.

    The maximum snow melt over a given number of days expressed in snow water equivalent.

    Parameters
    ----------
    snw : xarray.DataArray
        Snow amount (mass per area).
    window : int
        Number of days during which the melt is accumulated.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The maximum snow melt over a given number of days for each period. [mass/area].
    """
    # Compute change in SWE. Set melt as a positive change.
    dsnw = snw.diff(dim="time") * -1

    # Sum over window
    agg = dsnw.rolling(time=window).sum()

    # Max over period
    out = agg.resample(time=freq).max(dim="time")
    out.attrs["units"] = snw.units
    return out


@declare_units(snw="[mass]/[area]", pr="[precipitation]")
def melt_and_precip_max(
    snw: xarray.DataArray, pr: xarray.DataArray, window: int = 3, freq: str = "YS-JUL"
) -> xarray.DataArray:
    """
    Maximum snow melt and precipitation.

    The maximum snow melt plus precipitation over a given number of days expressed in snow water equivalent.

    Parameters
    ----------
    snw : xarray.DataArray
        Snow amount (mass per area).
    pr : xarray.DataArray
        Daily precipitation flux.
    window : int
        Number of days during which the water input is accumulated.
    freq : str
        Resampling frequency.

    Returns
    -------
    xarray.DataArray
        The maximum snow melt plus precipitation over a given number of days for each period. [mass/area].
    """
    # Compute change in SWE. Set melt as a positive change.
    dsnw = snw.diff(dim="time") * -1

    # Add precipitation total
    total = rate2amount(pr) + dsnw

    # Sum over window
    agg = total.rolling(time=window).sum()

    # Max over period
    out = agg.resample(time=freq).max(dim="time")
    out.attrs["units"] = snw.units
    return out


@declare_units(
    gwl="[length]",
    params="[]",
)
def standardized_groundwater_index(
    gwl: xarray.DataArray,
    freq: str | None = "MS",
    window: int = 1,
    dist: str | rv_continuous = "genextreme",
    method: str = "ML",
    fitkwargs: dict | None = None,
    cal_start: DateStr | None = None,
    cal_end: DateStr | None = None,
    params: Quantified | None = None,
    **indexer,
) -> xarray.DataArray:
    r"""
    Standardized Groundwater Index (SGI).

    Parameters
    ----------
    gwl : xarray.DataArray
        Groundwater head level.
    freq : str, optional
        Resampling frequency. A monthly or daily frequency is expected. Option `None` assumes
        that the desired resampling has already been applied input dataset and will skip the resampling step.
    window : int
        Averaging window length relative to the resampling frequency. For example, if `freq="MS"`,
        i.e. a monthly resampling, the window is an integer number of months.
    dist : {"gamma", "genextreme", "lognorm"} or `rv_continuous`
        Name of the univariate distribution, or a callable `rv_continuous` (see :py:mod:`scipy.stats`).
    method : {"APP", "ML", "PWM"}
        Name of the fitting method, such as `ML` (maximum likelihood), `APP` (approximate).
        The approximate method uses a deterministic function that does not involve any optimization,
        which can be sensitive to noise. `PWM` should be used with a `lmoments3` distribution.
    fitkwargs : dict, optional
        Kwargs passed to ``xclim.indices.stats.fit`` used to impose values of certain parameters (`floc`, `fscale`).
    cal_start : DateStr, optional
        Start date of the calibration period. A `DateStr` is expected, that is a `str` in format `"YYYY-MM-DD"`.
        Default option `None` means that the calibration period begins at the start of the input dataset.
    cal_end : DateStr, optional
        End date of the calibration period. A `DateStr` is expected, that is a `str` in format `"YYYY-MM-DD"`.
        Default option `None` means that the calibration period finishes at the end of the input dataset.
    params : xarray.DataArray, optional
        Fit parameters.
        The `params` can be computed using ``xclim.indices.stats.standardized_index_fit_params`` in advance.
        The output can be given here as input, and it overrides other options.
    **indexer : Indexer
        Indexing parameters to compute the indicator on a temporal subset of the data.
        It accepts the same arguments as :py:func:`xclim.indices.generic.select_time`.

    Returns
    -------
    xarray.DataArray, [unitless]
        Standardized Groundwater Index.

    See Also
    --------
    xclim.indices._agro.standardized_precipitation_index : Standardized Precipitation Index.
    xclim.indices.stats.standardized_index : Standardized Index.
    xclim.indices.stats.standardized_index_fit_params : Standardized Index Fit Params.

    Notes
    -----
    * N-month SGI / N-day SGI is determined by choosing the `window = N` and the appropriate frequency `freq`.
    * Supported statistical distributions are: ["gamma", "genextreme", "lognorm"].
    * If `params` is provided, it overrides the `cal_start`, `cal_end`, `freq`, `window`, `dist`, `method` options.
    * "APP" method only supports two-parameter distributions. Parameter `loc` needs to be fixed to use method "APP".

    References
    ----------
    :cite:cts:`bloomfield_2013`

    Examples
    --------
    >>> from datetime import datetime
    >>> from xclim.indices import standardized_groundwater_index
    >>> ds = xr.open_dataset(path_to_gwl_file)
    >>> gwl = ds.gwl
    >>> cal_start, cal_end = "1980-05-01", "1982-06-01"
    >>> sgi_3 = standardized_groundwater_index(
    ...     gwl,
    ...     freq="MS",
    ...     window=3,
    ...     dist="gamma",
    ...     method="ML",
    ...     cal_start=cal_start,
    ...     cal_end=cal_end,
    ... )  # Computing SGI-3 months using a Gamma distribution for the fit
    >>> # Fitting parameters can also be obtained first, then reused as input.
    >>> from xclim.indices.stats import standardized_index_fit_params
    >>> params = standardized_index_fit_params(
    ...     gwl.sel(time=slice(cal_start, cal_end)),
    ...     freq="MS",
    ...     window=3,
    ...     dist="gamma",
    ...     method="ML",
    ... )  # First getting params
    >>> sgi_3 = standardized_groundwater_index(gwl, params=params)
    """
    fitkwargs = fitkwargs or {}

    dist_methods = {
        "gamma": ["ML", "APP"],
        "genextreme": ["ML", "APP"],
        "lognorm": ["ML", "APP"],
    }
    if isinstance(dist, str):
        if dist in dist_methods:
            if method not in dist_methods[dist]:
                raise NotImplementedError(f"{method} method is not implemented for {dist} distribution")
        else:
            raise NotImplementedError(f"{dist} distribution is not yet implemented.")

    zero_inflated = False
    sgi = standardized_index(
        gwl,
        freq=freq,
        window=window,
        dist=dist,
        method=method,
        zero_inflated=zero_inflated,
        fitkwargs=fitkwargs,
        cal_start=cal_start,
        cal_end=cal_end,
        params=params,
        **indexer,
    )

    return sgi


@declare_units(q="[discharge]")
def flow_index(q: xarray.DataArray, p: float = 0.95) -> xarray.DataArray:
    """
    Flow index.

    Calculate the pth percentile of daily streamflow normalized by the median flow.

    Parameters
    ----------
    q : xarray.DataArray
        Daily streamflow data.
    p : float
        Percentile for calculating the flow index, between 0 and 1. Default of 0.95 is for high flows.

    Returns
    -------
    xarray.DataArray
        Normalized Qp, which is the p th percentile of daily streamflow normalized by the median flow.

    References
    ----------
    :cite:cts:`Clausen2000`
    """
    qp = q.quantile(p, dim="time")
    q_median = q.median(dim="time")
    out = qp / q_median
    out.attrs["units"] = "1"
    return out


@declare_units(q="[discharge]")
def high_flow_frequency(q: xarray.DataArray, threshold_factor: int = 9, freq: str = "YS-OCT") -> xarray.DataArray:
    """
    High flow frequency.

    Calculate the number of days in a given period with flows greater than a specified threshold, given as a
    multiple of the median flow. By default, the period is the water year starting on 1st October and ending on
    30th September, as commonly defined in North America.

    Parameters
    ----------
    q : xarray.DataArray
        Daily streamflow data.
    threshold_factor : int
        Factor by which the median flow is multiplied to set the high flow threshold, default is 9.
    freq : str, optional
        Resampling frequency, default is 'YS-OCT' for water year starting in October and ending in September.

    Returns
    -------
    xarray.DataArray
        Number of high flow days.

    References
    ----------
    :cite:cts:`addor2018,Clausen2000`
    """
    median_flow = q.median(dim="time")
    threshold = threshold_factor * median_flow
    out = threshold_count(q, ">", threshold, freq=freq)
    return to_agg_units(out, q, "count", deffreq="D")


@declare_units(q="[discharge]")
def low_flow_frequency(q: xarray.DataArray, threshold_factor: float = 0.2, freq: str = "YS-OCT") -> xarray.DataArray:
    """
    Low flow frequency.

    Calculate the number of days in a given period with flows lower than a specified threshold, given by a fraction
    of the mean flow. By default, the period is the water year starting on 1st October and ending on 30th September,
    as commonly defined in North America.

    Parameters
    ----------
    q : xarray.DataArray
        Daily streamflow data.
    threshold_factor : float
        Factor by which the mean flow is multiplied to set the low flow threshold, default is 0.2.
    freq : str, optional
        Resampling frequency, default is 'YS-OCT' for water year starting in October and ending in September.

    Returns
    -------
    xarray.DataArray
        Number of low flow days.

    References
    ----------
    :cite:cts:`Olden2003`
    """
    mean_flow = q.mean(dim="time")
    threshold = threshold_factor * mean_flow
    out = threshold_count(q, "<", threshold, freq=freq)
    return to_agg_units(out, q, "count", deffreq="D")


@declare_units(pr="[precipitation]")
def antecedent_precipitation_index(pr: xarray.DataArray, window: int = 7, p_exp: float = 0.935) -> xarray.DataArray:
    """
    Antecedent Precipitation Index.

    Calculate the running weighted sum of daily precipitation values given a window and weighting exponent.
    This index serves as an indicator for soil moisture.

    Parameters
    ----------
    pr : xarray.DataArray
        Daily precipitation data.
    window : int
        Window for the days of precipitation data to be weighted and summed, default is 7.
    p_exp : float
        Weighting exponent, default is 0.935.

    Returns
    -------
    xarray.DataArray
        Antecedent Precipitation Index.

    References
    ----------
    :cite:cts:`schroter2015,li2021`
    """
    pr = rate2amount(pr)
    pr = convert_units_to(pr, "mm", context="hydro")
    weights = xarray.DataArray(
        list(reversed([p_exp ** (idx - 1) for idx in range(1, window + 1)])),
        dims="window_dim",
    )
    out = pr.rolling(time=window).construct("window_dim").dot(weights)
    out.attrs["units"] = "mm"
    return out



#!/usr/bin/env python
# Test for utils
from __future__ import annotations

import numpy as np
import xarray as xr

from xclim.core.utils import _chunk_like, ensure_chunk_size, nan_calc_percentiles
from xclim.testing.helpers import test_timeseries as _test_timeseries


def test_ensure_chunk_size():
    da = xr.DataArray(np.zeros((20, 21, 20)), dims=("x", "y", "z"))

    out = ensure_chunk_size(da, x=10, y=-1)

    assert da is out

    dac = da.chunk({"x": (1,) * 20, "y": (10, 10, 1), "z": (10, 10)})

    out = ensure_chunk_size(dac, x=3, y=5, z=-1)

    assert out.chunks[0] == (3, 3, 3, 3, 3, 5)
    assert out.chunks[1] == (10, 11)
    assert out.chunks[2] == (20,)


class TestNanCalcPercentiles:
    def test_calc_perc_type7(self):
        # Example array from: https://en.wikipedia.org/wiki/Percentile#The_nearest-rank_method
        arr = np.asarray([15.0, 20.0, 35.0, 40.0, 50.0])
        res = nan_calc_percentiles(arr, percentiles=[40.0], alpha=1, beta=1)
        # The expected is from R `quantile(arr, probs=c(0.4), type=7)`
        assert res[()] == 29

    def test_calc_perc_type8(self):
        # Example array from: https://en.wikipedia.org/wiki/Percentile#The_nearest-rank_method
        arr = np.asarray([[15.0, 20.0, 35.0, 40.0, 50.0], [15.0, 20.0, 35.0, 40.0, 50.0]])
        res = nan_calc_percentiles(
            arr,
            percentiles=[40.0],
            alpha=1.0 / 3.0,
            beta=1.0 / 3.0,
        )
        # The expected is from R `quantile(arr, probs=c(0.4), type=8)`
        assert np.all(res[0][0] == 27)
        assert np.all(res[0][1] == 27)

    def test_calc_perc_2d(self):
        # Example array from: https://en.wikipedia.org/wiki/Percentile#The_nearest-rank_method
        arr = np.asarray([[15.0, 20.0, 35.0, 40.0, 50.0], [15.0, 20.0, 35.0, 40.0, 50.0]])
        res = nan_calc_percentiles(arr, percentiles=[40.0])
        # The expected is from R ` quantile(c(15.0, 20.0, 35.0, 40.0, 50.0), probs=0.4)`
        assert np.all(res[0][0] == 29)
        assert np.all(res[0][1] == 29)

    def test_calc_perc_nan(self):
        arr = np.asarray([np.nan])
        res = nan_calc_percentiles(arr, percentiles=[50.0])
        assert np.isnan(res)

    def test_calc_perc_empty(self):
        arr = np.asarray([])
        res = nan_calc_percentiles(arr)
        assert np.isnan(res)

    def test_calc_perc_partial_nan(self):
        arr = np.asarray([np.nan, 41.0, 41.0, 43.0, 43.0])
        res = nan_calc_percentiles(arr, percentiles=[50.0], alpha=1 / 3.0, beta=1 / 3.0)
        # The expected is from R `quantile(arr, 0.5, type=8, na.rm = TRUE)`
        # Note that scipy mquantiles would give a different result here
        assert res[()] == 42.0


def test_chunk_like():
    da = _test_timeseries(
        np.zeros(
            100,
        ),
        "tas",
    )
    da = xr.concat([da] * 10, xr.DataArray(np.arange(10), dims=("lat",), name="lat"))

    assert isinstance(da.lat.variable, xr.core.variable.IndexVariable)
    t, la = _chunk_like(da.time, da.lat, chunks={"time": 10, "lat": 1})
    assert t.chunks[0] == tuple([10] * 10)
    assert la.chunks[0] == tuple([1] * 10)



from __future__ import annotations

import numpy as np
import pytest

from xclim import atmos, convert


class TestWindSpeedIndicators:
    test_data = "ERA5/daily_surface_cancities_1990-1993.nc"

    def test_calm_windy_days(self, open_dataset):
        with open_dataset(self.test_data) as ds:
            sfcwind, _ = convert.wind_speed_from_vector(ds.uas, ds.vas, calm_wind_thresh="0 m/s")
            calm = atmos.calm_days(sfcwind, thresh="5 m/s")
            windy = atmos.windy_days(sfcwind, thresh="5 m/s")
            c = sfcwind.resample(time="MS").count()
            np.testing.assert_array_equal(calm + windy, c)


class TestSfcWind:
    test_data = "ERA5/daily_surface_cancities_1990-1993.nc"

    @pytest.mark.parametrize(
        "metric",
        ["mean", "max", "min"],
    )
    def test_sfcWind(self, open_dataset, metric):
        with open_dataset(self.test_data) as ds:
            sfcWind, _ = convert.wind_speed_from_vector(ds.uas, ds.vas)
            sfcWind_calculated = getattr(atmos, f"sfcWind_{metric}")(sfcWind)

            resample = sfcWind.resample(time="YS")
            c = getattr(resample, metric)()
            np.testing.assert_array_equal(sfcWind_calculated, c)


class TestSfcWindMax:
    test_data = "ERA5/daily_surface_cancities_1990-1993.nc"

    @pytest.mark.parametrize(
        "metric",
        ["mean", "max", "min"],
    )
    def test_sfcWindmax(self, open_dataset, metric):
        with open_dataset(self.test_data) as ds:
            sfcWind, _ = convert.wind_speed_from_vector(ds.uas, ds.vas)
            sfcWindmax_calculated = getattr(atmos, f"sfcWindmax_{metric}")(sfcWind)

            resample = sfcWind.resample(time="YS")
            c = getattr(resample, metric)()
            np.testing.assert_array_equal(sfcWindmax_calculated, c)



import numpy as np
import pytest

from xclim import atmos, set_options
from xclim.indices.helpers import make_hourly_temperature

K2C = 273.16


class TestLateFrostDays:
    def test_late_frost_days(self, atmosds):
        tasmin = atmosds.tasmin

        # Expected values
        exp = [2, 9, 72, 24, 0]

        lfd = atmos.late_frost_days(tasmin, date_bounds=("04-01", "06-30"))

        np.testing.assert_allclose(lfd.isel(time=0), exp, rtol=1e-03)


def test_high_precip_low_temp(pr_series, tasmin_series):
    pr = np.zeros(365)
    pr[1:3] = [1, 2]
    pr = pr_series(pr, start="1999-01-01")

    tas = np.zeros(365)
    tas[2:4] = [1, 1]
    tas += K2C
    tas = tasmin_series(tas, start="1999-01-01")

    out = atmos.high_precip_low_temp(pr, tas, pr_thresh="1 kg m-2 s-1", tas_thresh="1 C")
    np.testing.assert_array_equal(out, [1])


class TestDrynessIndex:
    def test_simple(self, atmosds):
        ds = atmosds.isel(location=3)

        pr = ds.pr
        evspsblpot = ds.evspsblpot

        di = atmos.dryness_index(pr, evspsblpot)
        np.testing.assert_allclose(di, np.array([13.355, 102.426, 65.576, 158.078]), rtol=1e-03)
        assert di.attrs["long_name"] == "Growing season humidity"

    def test_variable_initial_conditions(self, atmosds):
        ds = atmosds

        pr = ds.pr
        evspsblpot = ds.evspsblpot

        di = atmos.dryness_index(pr, evspsblpot)
        di_wet = atmos.dryness_index(pr, evspsblpot, wo="250 mm")
        di_dry = atmos.dryness_index(pr, evspsblpot, wo="100 mm")

        assert np.all(di_wet > di_dry)
        di_plus_50 = di + 50
        np.testing.assert_allclose(di_wet, di_plus_50, rtol=1e-03)
        di_minus_100 = di - 100
        np.testing.assert_allclose(di_dry, di_minus_100, rtol=1e-03)

        for value, array in {"200 mm": di, "250 mm": di_wet, "100 mm": di_dry}.items():
            assert array.attrs["long_name"] == "Growing season humidity"
            assert value in array.attrs["description"]


class TestChill:
    def test_chill_units(self, atmosds):
        tasmax = atmosds.tasmax
        tasmin = atmosds.tasmin
        tas = make_hourly_temperature(tasmin, tasmax)
        cu = atmos.chill_units(tas, date_bounds=("04-01", "06-30"))
        assert cu.attrs["units"] == "1"
        assert cu.name == "cu"
        assert cu.time.size == 4

        # Values are confirmed with chillR package although not an exact match
        # due to implementation details
        exp = [1546.5, 1344.0, 1162.0, 1457.5]
        np.testing.assert_allclose(cu.isel(location=0), exp, rtol=1e-03)

    @pytest.mark.parametrize("use_dask", [True, False])
    def test_chill_portions(self, atmosds, use_dask):
        pytest.importorskip("flox")
        tasmax = atmosds.tasmax
        tasmin = atmosds.tasmin
        tas = make_hourly_temperature(tasmin, tasmax)
        if use_dask:
            tas = tas.chunk(time=tas.time.size // 2, location=1)

        with set_options(resample_map_blocks=True):
            cp = atmos.chill_portions(tas, date_bounds=("09-01", "03-30"), freq="YS-JUL")

        assert cp.attrs["units"] == "1"
        assert cp.name == "cp"
        # Although its 4 years of data its 5 seasons starting in July
        assert cp.time.size == 5

        # Values are confirmed with chillR package although not an exact match
        # due to implementation details
        exp = [np.nan, 99.91534493, 92.5473925, 99.03177047, np.nan]
        np.testing.assert_allclose(cp.isel(location=0), exp, rtol=1e-03)


def test_water_cycle_intensity(pr_series, evspsbl_series):
    pr = pr_series(np.ones(31))
    evspsbl = evspsbl_series(np.ones(31))

    wci = atmos.water_cycle_intensity(pr=pr, evspsbl=evspsbl, freq="MS")
    np.testing.assert_allclose(wci, 2 * 60 * 60 * 24 * 31)



"""
Statistical downscaling and bias adjustment submodule.

This module is a placeholder for the `xclim.sdba` submodule, which has been split into its own package `xsdba`.
"""

import warnings

try:
    from xsdba import *  # pylint: disable=wildcard-import,unused-wildcard-import

    warnings.warn(
        "The `xclim.sdba` module has been split into its own package `xsdba`. "
        "Users are encouraged to use `xsdba` directly. "
        "For the time being, `xclim.sdba` will import `xsdba` to allow for API compatibility. "
        "This behaviour may change in the future. "
        "For more information, see: https://xsdba.readthedocs.io/en/stable/xclim_migration_guide.html"
    )
except ImportError as err:
    error_msg = (
        "The `xclim.sdba` module has been split into its own package: `xsdba`. "
        "Run `pip install xclim[extras]` or install `xsdba` via `pip` or `conda`. "
        "Users are encouraged to use `xsdba` directly. "
        "For the time being, `xclim.sdba` will import `xsdba` to allow for API compatibility. "
        "This behaviour may change in the future. "
        "For more information, see: https://xsdba.readthedocs.io/en/stable/xclim_migration_guide.html"
    )
    raise ImportError(error_msg) from err



"""Module for loading testing data."""

from __future__ import annotations

import logging
import warnings
from pathlib import Path
from typing import Any

import numpy as np
import pooch
import xarray as xr
from dask.callbacks import Callback

import xclim
from xclim.core import VARIABLES
from xclim.core.calendar import percentile_doy
from xclim.indices.converters import (
    longwave_upwelling_radiation_from_net_downwelling,
    shortwave_upwelling_radiation_from_net_downwelling,
)

logger = logging.getLogger("xclim")

__all__ = [
    "add_doctest_filepaths",
    "add_ensemble_dataset_objects",
    "add_example_file_paths",
    "assert_lazy",
    "generate_atmos",
    "test_timeseries",
]


def generate_atmos(
    nimbus: pooch.Pooch,
) -> dict[str, xr.DataArray]:
    """
    Create the `atmosds` synthetic testing dataset.

    Parameters
    ----------
    nimbus : pooch.Pooch
        The Pooch object to use for downloading the data.

    Returns
    -------
    dict[str, xr.DataArray]
        A dictionary of xarray DataArrays.
    """
    with xr.open_dataset(
        nimbus.fetch("ERA5/daily_surface_cancities_1990-1993.nc"),
        engine="h5netcdf",
    ) as ds:
        rsus = shortwave_upwelling_radiation_from_net_downwelling(ds.rss, ds.rsds)
        rlus = longwave_upwelling_radiation_from_net_downwelling(ds.rls, ds.rlds)
        tn10 = percentile_doy(ds.tasmin, per=10)
        t10 = percentile_doy(ds.tas, per=10)
        t90 = percentile_doy(ds.tas, per=90)
        tx90 = percentile_doy(ds.tasmax, per=90)

        ds = ds.assign(
            rsus=rsus,
            rlus=rlus,
            tn10=tn10,
            t10=t10,
            t90=t90,
            tx90=tx90,
        )

        # Create a file in a session-scoped temporary directory or the main cache
        atmos_file = Path(nimbus.path).joinpath("atmosds.nc")
        ds.to_netcdf(atmos_file, engine="h5netcdf")

    # Give access to dataset variables by name in namespace
    with xr.open_dataset(atmos_file, engine="h5netcdf") as ds:
        namespace = {f"{var}_dataset": ds[var] for var in ds.data_vars}
    return namespace


def add_ensemble_dataset_objects() -> dict[str, list[str]]:
    """
    Create a dictionary of xclim ensemble-related datasets to be patched into the xdoctest namespace.

    Returns
    -------
    dict[str, list[str]]
        A dictionary of xclim ensemble-related datasets.
    """
    namespace = {
        "nc_files_simple": [
            "EnsembleStats/BCCAQv2+ANUSPLIN300_ACCESS1-0_historical+rcp45_r1i1p1_1950-2100_tg_mean_YS.nc",
            "EnsembleStats/BCCAQv2+ANUSPLIN300_BNU-ESM_historical+rcp45_r1i1p1_1950-2100_tg_mean_YS.nc",
            "EnsembleStats/BCCAQv2+ANUSPLIN300_CCSM4_historical+rcp45_r1i1p1_1950-2100_tg_mean_YS.nc",
            "EnsembleStats/BCCAQv2+ANUSPLIN300_CCSM4_historical+rcp45_r2i1p1_1950-2100_tg_mean_YS.nc",
        ],
        "nc_files_extra": [
            "EnsembleStats/BCCAQv2+ANUSPLIN300_CNRM-CM5_historical+rcp45_r1i1p1_1970-2050_tg_mean_YS.nc"
        ],
    }
    namespace["nc_files"] = namespace["nc_files_simple"] + namespace["nc_files_extra"]
    return namespace


def add_example_file_paths() -> dict[str, str | list[xr.DataArray]]:
    """
    Create a dictionary of doctest-relevant datasets to be patched into the xdoctest namespace.

    Returns
    -------
    dict of str or dict of list of xr.DataArray
        A dictionary of doctest-relevant datasets.
    """
    namespace = {
        "path_to_ensemble_file": "EnsembleReduce/TestEnsReduceCriteria.nc",
        "path_to_gwl_file": "Raven/gwl_obs.nc",
        "path_to_pr_file": "NRCANdaily/nrcan_canada_daily_pr_1990.nc",
        "path_to_q_file": "Raven/q_sim.nc",
        "path_to_sfcWind_file": "ERA5/daily_surface_cancities_1990-1993.nc",
        "path_to_tas_file": "ERA5/daily_surface_cancities_1990-1993.nc",
        "path_to_tasmax_file": "NRCANdaily/nrcan_canada_daily_tasmax_1990.nc",
        "path_to_tasmin_file": "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc",
        "path_to_example_py": (
            Path(__file__).parent.parent.parent.parent / "docs" / "notebooks" / "example.py"
        ).as_posix(),
    }

    # For core.utils.load_module example
    sixty_years = xr.date_range("1990-01-01", "2049-12-31", freq="D")
    namespace["temperature_datasets"] = [
        xr.DataArray(
            12 * np.random.random_sample(sixty_years.size) + 273,
            coords={"time": sixty_years},
            name="tas",
            dims=("time",),
            attrs={
                "units": "K",
                "cell_methods": "time: mean within days",
                "standard_name": "air_temperature",
            },
        ),
        xr.DataArray(
            12 * np.random.random_sample(sixty_years.size) + 273,
            coords={"time": sixty_years},
            name="tas",
            dims=("time",),
            attrs={
                "units": "K",
                "cell_methods": "time: mean within days",
                "standard_name": "air_temperature",
            },
        ),
    ]
    return namespace


def add_doctest_filepaths() -> dict[str, Any]:
    """
    Overload some libraries directly into the xdoctest namespace.

    Returns
    -------
    dict[str, Any]
        A dictionary of xdoctest namespace objects.
    """
    namespace = {
        "np": np,
        "xclim": xclim,
        "tas": test_timeseries(np.random.rand(365) * 20 + 253.15, variable="tas"),
        "pr": test_timeseries(np.random.rand(365) * 5, variable="pr"),
    }
    return namespace


def test_timeseries(
    values,
    variable,
    start: str = "2000-07-01",
    units: str | None = None,
    freq: str = "D",
    as_dataset: bool = False,
    cftime: bool | None = None,
    calendar: str | None = None,
) -> xr.DataArray | xr.Dataset:
    """
    Create a generic timeseries object based on pre-defined dictionaries of existing variables.

    Parameters
    ----------
    values : np.ndarray
        The values of the DataArray.
    variable : str
        The name of the DataArray.
    start : str
        The start date of the time dimension. Default is "2000-07-01".
    units : str or None
        The units of the DataArray. Default is None.
    freq : str
        The frequency of the time dimension. Default is daily/"D".
    as_dataset : bool
        Whether to return a Dataset or a DataArray. Default is False.
    cftime : bool
        Whether to use cftime or not. Default is None, which uses cftime only for non-standard calendars.
    calendar : str or None
        Whether to use a calendar. If a calendar is provided, cftime is used.

    Returns
    -------
    xr.DataArray or xr.Dataset
        A DataArray or Dataset with time, lon and lat dimensions.
    """
    coords = xr.date_range(start, periods=len(values), freq=freq, calendar=calendar or "standard", use_cftime=cftime)

    if variable in VARIABLES:
        attrs = {a: VARIABLES[variable].get(a, "") for a in ["description", "standard_name", "cell_methods"]}
        attrs["units"] = VARIABLES[variable]["canonical_units"]

    else:
        warnings.warn(f"Variable {variable} not recognised. Attrs will not be filled.")
        attrs = {}

    if units is not None:
        attrs["units"] = units

    da = xr.DataArray(values, coords=[coords], dims="time", name=variable, attrs=attrs)

    if as_dataset:
        return da.to_dataset()
    return da


def _raise_on_compute(dsk: dict):
    """
    Raise an AssertionError mentioning the number triggered tasks.

    Parameters
    ----------
    dsk : dict
        The dask graph.

    Raises
    ------
    AssertionError
        If the dask computation is triggered.
    """
    raise AssertionError(f"Not lazy. Computation was triggered with a graph of {len(dsk)} tasks.")


assert_lazy = Callback(start=_raise_on_compute)
"""Context manager that raises an AssertionError if any dask computation is triggered."""
